from napari.benchmarks.utils import labeled_particles


def labeled_particles2d():
    seed = 275961054812084171
    labels, density, points = labeled_particles(
        (1024, 1024), seed=seed, return_density=True
    )

    return [
        (
            density.copy(),
            {'name': 'density', 'metadata': {'seed': seed}},
            'image',
        ),
        (
            labels.copy(),
            {'name': 'labels', 'metadata': {'seed': seed}},
            'labels',
        ),
        (
            points.copy(),
            {'name': 'points', 'metadata': {'seed': seed}},
            'points',
        ),
    ]


def labeled_particles3d():
    seed = 275961054812084171
    labels, density, points = labeled_particles(
        (256, 512, 512), seed=seed, return_density=True
    )

    return [
        (
            density.copy(),
            {'name': 'density', 'metadata': {'seed': seed}},
            'image',
        ),
        (
            labels.copy(),
            {'name': 'labels', 'metadata': {'seed': seed}},
            'labels',
        ),
        (
            points.copy(),
            {'name': 'points', 'metadata': {'seed': seed}},
            'points',
        ),
    ]



from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QComboBox,
    QHBoxLayout,
    QPushButton,
    QWidget,
)
from superqt import QLabeledDoubleSlider

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import qt_signals_blocked
from napari.layers import Image
from napari.layers.image._image_constants import VolumeDepiction
from napari.utils.action_manager import action_manager
from napari.utils.translations import trans


class PlaneNormalButtons(QWidget):
    """Qt buttons for controlling plane orientation.

    Attributes
    ----------
    x_button : qtpy.QtWidgets.QPushButton
        Button which orients a plane normal along the x axis.
    y_button : qtpy.QtWidgets.QPushButton
        Button which orients a plane normal along the y axis.
    z_button : qtpy.QtWidgets.QPushButton
        Button which orients a plane normal along the z axis.
    oblique_button : qtpy.QtWidgets.QPushButton
        Button which orients a plane normal along the camera view direction.
    """

    def __init__(self, parent=None) -> None:
        super().__init__(parent=parent)
        self.setLayout(QHBoxLayout())
        self.layout().setSpacing(2)
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.x_button = QPushButton('x')
        self.y_button = QPushButton('y')
        self.z_button = QPushButton('z')
        self.oblique_button = QPushButton(trans._('oblique'))
        action_manager.bind_button(
            'napari:orient_plane_normal_along_z',
            self.z_button,
        )
        action_manager.bind_button(
            'napari:orient_plane_normal_along_y',
            self.y_button,
        )
        action_manager.bind_button(
            'napari:orient_plane_normal_along_x',
            self.x_button,
        )
        action_manager.bind_button(
            'napari:orient_plane_normal_along_view_direction_no_gen',
            self.oblique_button,
        )

        self.layout().addWidget(self.x_button)
        self.layout().addWidget(self.y_button)
        self.layout().addWidget(self.z_button)
        self.layout().addWidget(self.oblique_button)


class QtDepictionControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the layer depection
    and plane value attributes and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Image
        An instance of a napari Image layer.

    Attributes
    ----------
    depiction_combobox : qtpy.QtWidgets.QComboBox
        ComboBox controlling current depiction value of the layer.
    depiction_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the depiction value chooser widget.
    plane_normal_buttons : PlaneNormalButtons
        Buttons controlling plane normal orientation when the `plane` depiction value is choosed.
    plane_normal_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the plane normal value chooser widget.
    plane_thickness_slider : superqt.QLabeledDoubleSlider
        Slider controlling plane normal thickness when the `plane` depiction value is choosed.
    plane_thickness_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the plane normal thickness value chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Image) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.depiction.connect(self._on_depiction_change)
        self._layer.plane.events.thickness.connect(
            self._on_plane_thickness_change
        )

        # Setup widgets
        self.depiction_combobox = QComboBox(parent)
        depiction_options = [d.value for d in VolumeDepiction]
        self.depiction_combobox.addItems(depiction_options)
        index = self.depiction_combobox.findText(
            self._layer.depiction, Qt.MatchFlag.MatchFixedString
        )
        self.depiction_combobox.setCurrentIndex(index)
        self.depiction_combobox.currentTextChanged.connect(
            self.change_depiction
        )
        self.depiction_label = QtWrappedLabel(trans._('depiction:'))

        # plane controls
        self.plane_normal_buttons = PlaneNormalButtons(parent)
        self.plane_normal_label = QtWrappedLabel(trans._('plane normal:'))

        self.plane_thickness_slider = QLabeledDoubleSlider(
            Qt.Orientation.Horizontal, parent
        )
        self.plane_thickness_slider.setFocusPolicy(Qt.NoFocus)
        self.plane_thickness_slider.setMinimum(1)
        self.plane_thickness_slider.setMaximum(50)
        self.plane_thickness_slider.setValue(self._layer.plane.thickness)
        self.plane_thickness_slider.valueChanged.connect(
            self.change_plane_thickness
        )
        self.plane_thickness_label = QtWrappedLabel(
            trans._('plane thickness:')
        )

    def change_depiction(self, text: str) -> None:
        self._layer.depiction = text
        self._update_plane_parameter_visibility()

    def change_plane_thickness(self, value: float) -> None:
        self._layer.plane.thickness = value

    def _on_depiction_change(self) -> None:
        """Receive layer model depiction change event and update combobox."""
        with qt_signals_blocked(self.depiction_combobox):
            index = self.depiction_combobox.findText(
                self._layer.depiction, Qt.MatchFlag.MatchFixedString
            )
            self.depiction_combobox.setCurrentIndex(index)
            self._update_plane_parameter_visibility()

    def _on_plane_thickness_change(self) -> None:
        with self._layer.plane.events.blocker():
            self.plane_thickness_slider.setValue(self._layer.plane.thickness)

    def _on_display_change_hide(self) -> None:
        self.depiction_combobox.hide()
        self.depiction_label.hide()

    def _on_display_change_show(self) -> None:
        self.depiction_combobox.show()
        self.depiction_label.show()

    def _update_plane_parameter_visibility(self) -> None:
        """Hide plane rendering controls if they aren't needed."""
        depiction = VolumeDepiction(self._layer.depiction)
        # TODO: Better way to handle the ndisplay value?
        visible = (
            depiction == VolumeDepiction.PLANE
            and self.parent().ndisplay == 3
            and self._layer.ndim >= 3
        )
        self.plane_normal_buttons.setVisible(visible)
        self.plane_normal_label.setVisible(visible)
        self.plane_thickness_slider.setVisible(visible)
        self.plane_thickness_label.setVisible(visible)

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [
            (self.depiction_label, self.depiction_combobox),
            (self.plane_normal_label, self.plane_normal_buttons),
            (self.plane_thickness_label, self.plane_thickness_slider),
        ]



import numpy as np
from qtpy.QtWidgets import (
    QDoubleSpinBox,
    QWidget,
)

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import attr_to_settr
from napari.layers import Vectors
from napari.utils.translations import trans


class QtWidthSpinBoxControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the layer edge width
    attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Vectors
        An instance of a napari Vectors layer.

    Attributes
    ----------
    width_spinbox : qtpy.QtWidgets.QDoubleSpinBox
        Spin box widget controlling edge width of vectors.
    width_spinbox_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the edge width of vectors chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Vectors) -> None:
        super().__init__(parent, layer)
        # Setup widgets
        # line width in pixels
        self.width_spinbox = QDoubleSpinBox()
        self.width_spinbox.setKeyboardTracking(False)
        self.width_spinbox.setSingleStep(0.1)
        self.width_spinbox.setMinimum(0.01)
        self.width_spinbox.setMaximum(np.inf)
        self.width_spinbox.setValue(self._layer.edge_width)
        self.width_spinbox.valueChanged.connect(self.change_width)
        self._callbacks.append(
            attr_to_settr(
                self._layer,
                'edge_width',
                self.width_spinbox,
                'setValue',
            )
        )
        self.width_spinbox_label = QtWrappedLabel(trans._('width:'))

    def change_width(self, value) -> None:
        """Change edge line width of vectors on the layer model.
        Parameters
        ----------
        value : float
            Line width of vectors.
        """
        self._layer.edge_width = value
        self.width_spinbox.clearFocus()
        # TODO: Check other way to give focus without calling parent
        self.parent().setFocus()

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.width_spinbox_label, self.width_spinbox)]


class QtLengthSpinBoxControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the layer length of vectors
    attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Vectors
        An instance of a napari Vectors layer.

    Attributes
    ----------
    length_spinbox : qtpy.QtWidgets.QDoubleSpinBox
        Spinbox widget controlling line length of vectors.
        Multiplicative factor on projections for length of all vectors.
    length_spinbox_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for line length of vectors chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Vectors) -> None:
        super().__init__(parent, layer)
        # Setup widgets
        # line length
        self.length_spinbox = QDoubleSpinBox()
        self.length_spinbox.setKeyboardTracking(False)
        self.length_spinbox.setSingleStep(0.1)
        self.length_spinbox.setValue(self._layer.length)
        self.length_spinbox.setMinimum(0.1)
        self.length_spinbox.setMaximum(np.inf)
        self.length_spinbox.valueChanged.connect(self.change_length)
        self._callbacks.append(
            attr_to_settr(
                self._layer,
                'length',
                self.length_spinbox,
                'setValue',
            )
        )
        self.length_spinbox_label = QtWrappedLabel(trans._('length:'))

    def change_length(self, value: float) -> None:
        """Change length of vectors on the layer model.
        Multiplicative factor on projections for length of all vectors.
        Parameters
        ----------
        value : float
            Length of vectors.
        """
        self._layer.length = value
        self.length_spinbox.clearFocus()
        # TODO: Check other way to give focus without calling parent
        self.parent().setFocus()

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.length_spinbox_label, self.length_spinbox)]



from napari.components._viewer_constants import CanvasPosition
from napari.components.overlays.base import CanvasOverlay
from napari.utils.color import ColorValue


class ColorBarOverlay(CanvasOverlay):
    """ColorBar legend overlay.

    Attributes
    ----------
    position : CanvasPosition
        The position of the overlay in the canvas.
    visible : bool
        If the overlay is visible or not.
    opacity : float
        The opacity of the overlay. 0 is fully transparent.
    order : int
        The rendering order of the overlay: lower numbers get rendered first.
    """

    color: ColorValue | None = None
    size: tuple[float, float] = 25, 150
    tick_length: float = 5
    font_size: float = 10
    position: CanvasPosition = CanvasPosition.TOP_RIGHT



from qtpy.QtCore import Qt
from qtpy.QtWidgets import QComboBox, QWidget

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import qt_signals_blocked
from napari._qt.widgets.qt_color_swatch import QColorSwatchEdit
from napari.layers import Vectors
from napari.layers.utils._color_manager_constants import ColorMode
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtEdgeColorFeatureControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the current edge
    color, color mode and color feature selection from the layer attributes and
    Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Vectors
        An instance of a napari Vectors layer.

    Attributes
    ----------
    color_mode_combobox : qtpy.QtWidgets.QComboBox
        Dropdown to select the edge color mode.
    color_mode_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the current selected edge_color_mode chooser widget.
    edge_color_edit : qtpy.QtWidgets.QSlider
        ColorSwatchEdit controlling current edge color of the layer.
    edge_color_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the current edge color chooser widget.
    color_feature_box : qtpy.QtWidgets.QComboBox
        Dropdown to select the feature for mapping edge_color.
    edge_feature_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the color_feature_box chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Vectors) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.edge_color_mode.connect(
            self._on_edge_color_mode_change
        )
        self._layer.events.edge_color.connect(self._on_edge_color_change)

        # Setup widgets
        # dropdown to select the feature for mapping edge_color
        self.color_feature_box = QComboBox(parent)
        self.color_feature_box.currentTextChanged.connect(
            self.change_edge_color_feature
        )
        self.color_feature_box.addItems(self._layer.features.columns)
        self.edge_feature_label = QtWrappedLabel(trans._('edge feature:'))

        # vector direct color mode adjustment and widget
        self.edge_color_edit = QColorSwatchEdit(
            initial_color=self._layer.edge_color,
            tooltip=trans._(
                'Click to set current edge color',
            ),
        )
        connect_setattr(
            self.edge_color_edit.color_changed, self._layer, 'edge_color'
        )
        self.edge_color_label = QtWrappedLabel(trans._('edge color:'))
        self._on_edge_color_change()

        # dropdown to select the edge color mode
        self.color_mode_combobox = QComboBox(parent)
        color_modes = [e.value for e in ColorMode]
        self.color_mode_combobox.addItems(color_modes)
        self.color_mode_combobox.currentTextChanged.connect(
            self.change_edge_color_mode
        )
        self.color_mode_label = QtWrappedLabel(trans._('edge color mode:'))
        self._on_edge_color_mode_change()

    def change_edge_color_feature(self, feature: str):
        """Change edge_color feature of vectors on the layer model.

        Parameters
        ----------
        feature : str
            feature to map the edge color to
        """
        mode = self._layer.edge_color_mode
        try:
            self._layer.edge_color = feature
            self._layer.edge_color_mode = mode
        except TypeError:
            # if the selected feature is the wrong type for the current color mode
            # the color mode will be changed to the appropriate type, so we must update
            self._on_edge_color_mode_change()
            raise

    def change_edge_color_mode(self, mode: str):
        """Change edge color mode of vectors on the layer model.

        Parameters
        ----------
        mode : str
            Edge color for vectors. Must be: 'direct', 'cycle', or 'colormap'
        """
        old_mode = self._layer.edge_color_mode
        with self._layer.events.edge_color_mode.blocker():
            try:
                self._layer.edge_color_mode = mode
                self._update_edge_color_gui(mode)

            except ValueError:
                # if the color mode was invalid, revert to the old mode (layer and GUI)
                self._layer.edge_color_mode = old_mode
                self.color_mode_combobox.setCurrentText(old_mode)
                raise

    def _on_edge_color_mode_change(self):
        """Receive layer model edge color mode change event & update dropdown."""
        if not hasattr(self, 'color_mode_combobox'):
            # Ignore early events i.e when widgets haven't been created yet.
            return

        with qt_signals_blocked(self.color_mode_combobox):
            mode = self._layer._edge.color_mode
            index = self.color_mode_combobox.findText(
                mode, Qt.MatchFixedString
            )
            self.color_mode_combobox.setCurrentIndex(index)

            self._update_edge_color_gui(mode)

    def _on_edge_color_change(self):
        """Receive layer model edge color  change event & update dropdown."""
        if (
            self._layer._edge.color_mode == ColorMode.DIRECT
            and len(self._layer.data) > 0
        ):
            with qt_signals_blocked(self.edge_color_edit):
                self.edge_color_edit.setColor(self._layer.edge_color[0])
        elif self._layer._edge.color_mode in (
            ColorMode.CYCLE,
            ColorMode.COLORMAP,
        ):
            with qt_signals_blocked(self.color_feature_box):
                prop = self._layer._edge.color_properties.name
                index = self.color_feature_box.findText(
                    prop, Qt.MatchFixedString
                )
                self.color_feature_box.setCurrentIndex(index)

    def _update_edge_color_gui(self, mode: str):
        """Update the GUI element associated with edge_color.
        This is typically used when edge_color_mode changes

        Parameters
        ----------
        mode : str
            The new edge_color mode the GUI needs to be updated for.
            Should be: 'direct', 'cycle', 'colormap'
        """
        if mode in {'cycle', 'colormap'}:
            self.edge_color_edit.setHidden(True)
            self.edge_color_label.setHidden(True)
            self.color_feature_box.setHidden(False)
            self.edge_feature_label.setHidden(False)

        elif mode == 'direct':
            self.edge_color_edit.setHidden(False)
            self.edge_color_label.setHidden(False)
            self.color_feature_box.setHidden(True)
            self.edge_feature_label.setHidden(True)

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [
            (self.color_mode_label, self.color_mode_combobox),
            (self.edge_color_label, self.edge_color_edit),
            (self.edge_feature_label, self.color_feature_box),
        ]



from typing import TYPE_CHECKING

from napari._qt.layer_controls.qt_layer_controls_base import QtLayerControls
from napari._qt.layer_controls.widgets import (
    QtFaceColorControl,
    QtOutSliceCheckBoxControl,
    QtProjectionModeControl,
    QtTextVisibilityControl,
)
from napari._qt.layer_controls.widgets._points import (
    QtBorderColorControl,
    QtCurrentSizeSliderControl,
    QtSymbolComboBoxControl,
)
from napari._qt.widgets.qt_mode_buttons import QtModePushButton
from napari.layers.points._points_constants import Mode
from napari.utils.action_manager import action_manager
from napari.utils.events import disconnect_events
from napari.utils.translations import trans

if TYPE_CHECKING:
    import napari.layers


class QtPointsControls(QtLayerControls):
    """Qt view and controls for the napari Points layer.

    Parameters
    ----------
    layer : napari.layers.Points
        An instance of a napari Points layer.

    Attributes
    ----------
    _border_color_control : napari._qt.layer_controls.widgets._points.QtBorderColorControl
        Widget to handle point's border color.
    _current_size_slider_control : napari._qt.layer_controls.widgets._points.QtCurrentSizeSliderControl
        Widget that wraps slider controlling size of points.
    _face_color_control : napari._qt.layer_controls.widgets.QtFaceColorControl
        Widget to select display color for points faces.
    _out_slice_checkbox_control : napari._qt.layer_controls.widgets.QtOutSliceCheckBoxControl
        Widget that wraps a checkbox to indicate whether to render out of slice.
    _projection_mode_control : napari._qt.layer_controls.widgets.QtProjectionModeControl
        Widget that wraps dropdown menu to select the projection mode for the layer.
    _symbol_combobox_control : napari._qt.layer_controls.widgets._points.QtSymbolComboBoxControl
        Widget that wraps a dropdown list of symbol options for points markers.
    _text_visibility_control : napari._qt.layer_controls.widgets.QtTextVisibilityControl
        Widget that wraps a checkbox controlling if text on the layer is visible or not.
    addition_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to add points to layer.
    delete_button : qtpy.QtWidgets.QtModePushButton
        Button to delete points from layer.
    select_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to select points from layer.

    Raises
    ------
    ValueError
        Raise error if points mode is not recognized.
        Points mode must be one of: ADD, PAN_ZOOM, or SELECT.
    """

    layer: 'napari.layers.Points'
    MODE = Mode
    PAN_ZOOM_ACTION_NAME = 'activate_points_pan_zoom_mode'
    TRANSFORM_ACTION_NAME = 'activate_points_transform_mode'

    def __init__(self, layer) -> None:
        super().__init__(layer)

        # Setup buttons
        self.select_button = self._radio_button(
            layer,
            'select_points',
            Mode.SELECT,
            True,
            'activate_points_select_mode',
        )
        self.addition_button = self._radio_button(
            layer,
            'add_points',
            Mode.ADD,
            True,
            'activate_points_add_mode',
        )
        self.delete_button = QtModePushButton(
            layer,
            'delete_shape',
        )
        action_manager.bind_button(
            'napari:delete_selected_points', self.delete_button
        )
        self._EDIT_BUTTONS += (self.delete_button,)
        self._on_editable_or_visible_change()

        self.button_grid.addWidget(self.delete_button, 0, 3)
        self.button_grid.addWidget(self.addition_button, 0, 4)
        self.button_grid.addWidget(self.select_button, 0, 5)

        # Setup widgets controls
        self._projection_mode_control = QtProjectionModeControl(self, layer)
        self._add_widget_controls(self._projection_mode_control)
        self._current_size_slider_control = QtCurrentSizeSliderControl(
            self, layer
        )
        self._add_widget_controls(self._current_size_slider_control)
        self._symbol_combobox_control = QtSymbolComboBoxControl(self, layer)
        self._add_widget_controls(self._symbol_combobox_control)
        self._face_color_control = QtFaceColorControl(
            self,
            layer,
            tooltip=trans._(
                'Click to set the face color of currently selected points and any added afterwards.'
            ),
        )
        self._add_widget_controls(self._face_color_control)
        self._border_color_control = QtBorderColorControl(self, layer)
        self._add_widget_controls(self._border_color_control)
        self._text_visibility_control = QtTextVisibilityControl(self, layer)
        self._add_widget_controls(self._text_visibility_control)
        self._out_slice_checkbox_control = QtOutSliceCheckBoxControl(
            self, layer
        )
        self._add_widget_controls(self._out_slice_checkbox_control)

    def _on_mode_change(self, event):
        """Update ticks in checkbox widgets when points layer mode is changed.

        Available modes for points layer are:
        * ADD
        * SELECT
        * PAN_ZOOM
        * TRANSFORM

        Parameters
        ----------
        event : napari.utils.event.Event
            The napari event that triggered this method.

        Raises
        ------
        ValueError
            Raise error if event.mode is not ADD, PAN_ZOOM, TRANSFORM or SELECT.
        """
        super()._on_mode_change(event)

    def _on_ndisplay_changed(self):
        self.layer.editable = not (self.layer.ndim == 2 and self.ndisplay == 3)
        super()._on_ndisplay_changed()

    def close(self):
        """Disconnect events when widget is closing."""
        disconnect_events(self.layer.text.events, self)
        super().close()



from qtpy.QtWidgets import (
    QCheckBox,
    QWidget,
)

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import attr_to_settr, checked_to_bool
from napari.layers import Labels
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtPreserveLabelsCheckBoxControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the layer attribute to
    preserve existing labels and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Labels
        An instance of a napari Labels layer.

    Attributes
    ----------
    preserve_labels_checkbox : qtpy.QtWidgets.QCheckBox
        Checkbox to control if existing labels are preserved.
    preserve_labels_checkbox_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the layer should preserve labels chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Labels) -> None:
        super().__init__(parent, layer)
        # Setup widgets
        preserve_labels_cb = QCheckBox()
        preserve_labels_cb.setToolTip(
            trans._('Preserve existing labels while painting')
        )
        preserve_labels_cb.setChecked(self._layer.preserve_labels)
        self._callbacks.append(
            attr_to_settr(
                self._layer,
                'preserve_labels',
                preserve_labels_cb,
                'setChecked',
            )
        )
        connect_setattr(
            preserve_labels_cb.stateChanged,
            layer,
            'preserve_labels',
            convert_fun=checked_to_bool,
        )
        self.preserve_labels_checkbox = preserve_labels_cb

        self.preserve_labels_checkbox_label = QtWrappedLabel(
            trans._('preserve\nlabels:')
        )

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [
            (
                self.preserve_labels_checkbox_label,
                self.preserve_labels_checkbox,
            )
        ]



from qtpy.QtWidgets import (
    QComboBox,
    QWidget,
)

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import qt_signals_blocked
from napari.layers import Image
from napari.layers.image._image_constants import Interpolation
from napari.utils.translations import trans


class QtInterpolationComboBoxControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the layer interpolation
    mode attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Image
        An instance of a napari Image layer.

    Attributes
    ----------
    interpolation_combobox : qtpy.QtWidgets.QComboBox
        ComboBox controlling current shading value of the layer.
    interpolation_combobox_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the shading value chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Image) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.interpolation2d.connect(
            self._on_interpolation_change
        )
        self._layer.events.interpolation3d.connect(
            self._on_interpolation_change
        )

        # Setup widgets
        self.interpolation_combobox = QComboBox(parent)
        self.interpolation_combobox.currentTextChanged.connect(
            self.change_interpolation
        )
        self.interpolation_combobox.setToolTip(
            trans._(
                'Texture interpolation for display.\nnearest and linear are most performant.'
            )
        )

        self.interpolation_combobox_label = QtWrappedLabel(
            trans._('interpolation:')
        )

    def change_interpolation(self, text: str) -> None:
        """Change interpolation mode for image display.

        Parameters
        ----------
        text : str
            Interpolation mode used by vispy. Must be one of our supported
            modes:
            'bessel', 'bicubic', 'linear', 'blackman', 'catrom', 'gaussian',
            'hamming', 'hanning', 'hermite', 'kaiser', 'lanczos', 'mitchell',
            'nearest', 'spline16', 'spline36'
        """
        # TODO: Better way to handle the ndisplay value?
        if self.parent().ndisplay == 2:
            self._layer.interpolation2d = text
        else:
            self._layer.interpolation3d = text

    def _on_interpolation_change(self, event) -> None:
        """Receive layer interpolation change event and update dropdown menu.

        Parameters
        ----------
        event : napari.utils.event.Event
            The napari event that triggered this method.
        """
        interp_string = event.value.value

        with qt_signals_blocked(self.interpolation_combobox):
            if self.interpolation_combobox.findText(interp_string) == -1:
                self.interpolation_combobox.addItem(interp_string)
            self.interpolation_combobox.setCurrentText(interp_string)

    def _update_interpolation_combo(self, ndisplay: int) -> None:
        interp_names = [i.value for i in Interpolation.view_subset()]
        interp = (
            self._layer.interpolation2d
            if ndisplay == 2
            else self._layer.interpolation3d
        )
        with qt_signals_blocked(self.interpolation_combobox):
            self.interpolation_combobox.clear()
            self.interpolation_combobox.addItems(interp_names)
            self.interpolation_combobox.setCurrentText(interp)

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [
            (self.interpolation_combobox_label, self.interpolation_combobox)
        ]



from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QComboBox,
    QWidget,
)
from superqt import QLabeledDoubleSlider

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import attr_to_settr
from napari.layers.base._base_constants import BLENDING_TRANSLATIONS, Blending
from napari.layers.base.base import Layer
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans

# opaque and minimum blending do not support changing alpha (opacity)
NO_OPACITY_BLENDING_MODES = {str(Blending.MINIMUM), str(Blending.OPAQUE)}


class QtOpacityBlendingControls(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between opacity/blending
    layer attributes and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Layer
        An instance of a napari layer.

    Attributes
    ----------
    blend_combobox : qtpy.QtWidgets.QComboBox
        Dropdown widget to select blending mode of layer.
    blend_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the blending combobox widget.
    opacity_slider : superqt.QLabeledDoubleSlider
        Slider controlling opacity of the layer.
    opacity_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the opacity slider widget.
    """

    def __init__(self, parent: QWidget, layer: Layer) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.blending.connect(self._on_blending_change)

        # Setup widgets
        sld = QLabeledDoubleSlider(Qt.Orientation.Horizontal, parent=parent)
        sld.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        sld.setMinimum(0)
        sld.setMaximum(1)
        sld.setSingleStep(0.01)
        sld.setValue(self._layer.opacity)
        self.opacity_slider = sld
        connect_setattr(
            self.opacity_slider.valueChanged, self._layer, 'opacity'
        )
        self._callbacks.append(
            attr_to_settr(
                self._layer, 'opacity', self.opacity_slider, 'setValue'
            )
        )
        self.opacity_label = QtWrappedLabel(trans._('opacity:'))

        blend_combobox = QComboBox(parent)
        for index, (data, text) in enumerate(BLENDING_TRANSLATIONS.items()):
            data = data.value
            blend_combobox.addItem(text, data)
            if data == self._layer.blending:
                blend_combobox.setCurrentIndex(index)

        blend_combobox.currentTextChanged.connect(self.change_blending)
        self.blend_combobox = blend_combobox
        self.blend_label = QtWrappedLabel(trans._('blending:'))

        # opaque and minimum blending do not support changing alpha
        self.opacity_slider.setEnabled(
            self._layer.blending not in NO_OPACITY_BLENDING_MODES
        )
        self.opacity_label.setEnabled(
            self._layer.blending not in NO_OPACITY_BLENDING_MODES
        )

    def change_blending(self, text: str) -> None:
        """Change blending mode on the layer model.

        Parameters
        ----------
        text : str
            Name of blending mode, eg: 'translucent', 'additive', 'opaque'.
        """
        self._layer.blending = self.blend_combobox.currentData()
        # opaque and minimum blending do not support changing alpha
        self.opacity_slider.setEnabled(
            self._layer.blending not in NO_OPACITY_BLENDING_MODES
        )
        self.opacity_label.setEnabled(
            self._layer.blending not in NO_OPACITY_BLENDING_MODES
        )

        blending_tooltip = ''
        if self._layer.blending == Blending.MINIMUM:
            blending_tooltip = trans._(
                '`minimum` blending mode works best with inverted colormaps with a white background.',
            )
        self.blend_combobox.setToolTip(blending_tooltip)
        self._layer.help = blending_tooltip

    def _on_blending_change(self) -> None:
        """Receive layer model blending mode change event and update slider."""
        with self._layer.events.blending.blocker():
            self.blend_combobox.setCurrentIndex(
                self.blend_combobox.findData(self._layer.blending)
            )

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [
            (self.opacity_label, self.opacity_slider),
            (self.blend_label, self.blend_combobox),
        ]



from __future__ import annotations

import itertools
import warnings
from typing import TYPE_CHECKING

import numpy as np
import pint

from napari.layers import Image
from napari.layers.image._image_utils import guess_multiscale
from napari.utils.colormaps import CYMRGB, MAGENTA_GREEN, Colormap
from napari.utils.misc import ensure_iterable, ensure_sequence_of_iterables
from napari.utils.translations import trans

if TYPE_CHECKING:
    from napari.types import FullLayerData


def slice_from_axis(array, *, axis, element):
    """Take a single index slice from array using slicing.

    Equivalent to :func:`np.take`, but using slicing, which ensures that the
    output is a view of the original array.

    Parameters
    ----------
    array : NumPy or other array
        Input array to be sliced.
    axis : int
        The axis along which to slice.
    element : int
        The element along that axis to grab.

    Returns
    -------
    sliced : NumPy or other array
        The sliced output array, which has one less dimension than the input.
    """
    # Check if array is a zarr array and wrap it with dask
    # to keep lazy behavior and avoid loading into memory
    if hasattr(array, '__module__') and array.__module__.startswith('zarr'):
        import dask.array as da

        array = da.from_zarr(array)
        warnings.warn(
            trans._(
                'zarr array cannot be sliced lazily, converted to dask array.',
                deferred=True,
            )
        )

    slices = [slice(None) for i in range(array.ndim)]
    slices[axis] = element
    return array[tuple(slices)]


def split_channels(
    data: np.ndarray,
    channel_axis: int,
    **kwargs,
) -> list[FullLayerData]:
    """Split the data array into separate arrays along an axis.

    Keyword arguments will override any parameters altered or set in this
    function. Colormap, blending, or multiscale are set as follows if not
    overridden by a keyword:
    - colormap : (magenta, green) for 2 channels, (CYMRGB) for more than 2
    - blending : translucent for first channel, additive for others
    - multiscale : determined by layers.image._image_utils.guess_multiscale.

    Colormap, blending and multiscale will be set and returned in meta if not in kwargs.
    If any other key is not present in kwargs it will not be returned in the meta
    dictionary of the returned LaterData tuple. For example, if gamma is not in
    kwargs then meta will not have a gamma key.

    Parameters
    ----------
    data : array or list of array
    channel_axis : int
        Axis to split the image along.
    **kwargs : dict
        Keyword arguments will override the default image meta keys
        returned in each layer data tuple.

    Returns
    -------
    List of LayerData tuples: [(data: array, meta: Dict, type: str )]
    """

    # Determine if data is a multiscale
    multiscale = kwargs.get('multiscale')
    if not multiscale:
        multiscale, data = guess_multiscale(data)
        kwargs['multiscale'] = multiscale

    n_channels = (data[0] if multiscale else data).shape[channel_axis]
    # Use original blending mode or for multichannel use translucent for first channel then additive
    kwargs['blending'] = kwargs.get('blending') or ['translucent_no_depth'] + [
        'additive'
    ] * (n_channels - 1)
    kwargs.setdefault('colormap', None)
    # these arguments are *already* iterables in the single-channel case.
    iterable_kwargs = {
        'axis_labels',
        'scale',
        'translate',
        'contrast_limits',
        'metadata',
        'plane',
        'experimental_clipping_planes',
        'custom_interpolation_kernel_2d',
        'units',
    }

    # turn the kwargs dict into a mapping of {key: iterator}
    # so that we can use {k: next(v) for k, v in kwargs.items()} below
    for key, val in kwargs.items():
        if key == 'colormap' and val is None:
            if n_channels == 1:
                kwargs[key] = iter(['gray'])
            elif n_channels == 2:
                kwargs[key] = iter(MAGENTA_GREEN)
            else:
                kwargs[key] = itertools.cycle(CYMRGB)

        # make sure that iterable_kwargs are a *sequence* of iterables
        # for the multichannel case.  For example: if scale == (1, 2) &
        # n_channels = 3, then scale should == [(1, 2), (1, 2), (1, 2)]
        elif key in iterable_kwargs or (
            key == 'colormap' and isinstance(val, Colormap)
        ):
            kwargs[key] = iter(
                ensure_sequence_of_iterables(
                    val,
                    n_channels,
                    repeat_empty=True,
                    allow_none=True,
                )
            )
        elif key in ['rotate', 'shear'] or (
            key == 'affine' and isinstance(val, np.ndarray)
        ):
            # affine may be Affine or np.ndarray object that is not
            # iterable, but it is not now a problem as we use it only to warning
            # if a provided object is a sequence and channel_axis is not provided
            kwargs[key] = itertools.repeat(val, n_channels)
        else:
            kwargs[key] = iter(ensure_iterable(val))

    layerdata_list = []
    for i in range(n_channels):
        if multiscale:
            image = [
                slice_from_axis(data[j], axis=channel_axis, element=i)
                for j in range(len(data))
            ]
        else:
            image = slice_from_axis(data, axis=channel_axis, element=i)
        i_kwargs = {}
        for key, val in kwargs.items():
            try:
                i_kwargs[key] = next(val)
            except StopIteration as e:
                raise IndexError(
                    trans._(
                        "Error adding multichannel image with data shape {data_shape!r}.\nRequested channel_axis ({channel_axis}) had length {n_channels}, but the '{key}' argument only provided {i} values. ",
                        deferred=True,
                        data_shape=data.shape,
                        channel_axis=channel_axis,
                        n_channels=n_channels,
                        key=key,
                        i=i,
                    )
                ) from e

        layerdata: FullLayerData = (image, i_kwargs, 'image')
        layerdata_list.append(layerdata)

    return layerdata_list


def stack_to_images(stack: Image, axis: int, **kwargs) -> list[Image]:
    """Splits a single Image layer into a list layers along axis.

    Some image layer properties will be changed unless specified as an item in
    kwargs. Properties such as colormap and contrast_limits are set on individual
    channels. Properties will be changed as follows (unless overridden with a kwarg):
    - colormap : (magenta, green) for 2 channels, (CYMRGB) for more than 2
    - blending : additive
    - contrast_limits : min and max of the image

    All other properties, such as scale and translate will be propagated from the
    original stack, unless a keyword argument passed for that property.

    Parameters
    ----------
    stack : napari.layers.Image
        The image stack to be split into a list of image layers
    axis : int
        The axis to split along.

    Returns
    -------
    imagelist: list
        List of Image objects
    """

    data, meta, _ = stack.as_layer_data_tuple()

    for key in ('contrast_limits', 'colormap', 'blending'):
        del meta[key]

    name = stack.name
    num_dim = 3 if stack.rgb else stack.ndim

    if num_dim < 3:
        raise ValueError(
            trans._(
                'The image needs more than 2 dimensions for splitting',
                deferred=True,
            )
        )

    if axis >= num_dim:
        raise ValueError(
            trans._(
                "Can't split along axis {axis}. The image has {num_dim} dimensions",
                deferred=True,
                axis=axis,
                num_dim=num_dim,
            )
        )

    if kwargs.get('colormap'):
        kwargs['colormap'] = itertools.cycle(kwargs['colormap'])

    if meta['rgb']:
        if axis in [num_dim - 1, -1]:
            kwargs['rgb'] = False  # split channels as grayscale
        else:
            kwargs['rgb'] = True  # split some other axis, remain rgb
            meta['scale'].pop(axis)
            meta['translate'].pop(axis)
    else:
        kwargs['rgb'] = False
        meta['scale'].pop(axis)
        meta['translate'].pop(axis)

    meta['rotate'] = None
    meta['shear'] = None
    meta['affine'] = None
    meta['axis_labels'] = None
    meta['units'] = None

    meta.update(kwargs)
    imagelist = []
    layerdata_list = split_channels(data, axis, **meta)
    for i, tup in enumerate(layerdata_list):
        idata, imeta, _ = tup
        layer_name = f'{name} layer {i}'
        imeta['name'] = layer_name

        imagelist.append(Image(idata, **imeta))

    return imagelist


def split_rgb(stack: Image, with_alpha=False) -> list[Image]:
    """Split RGB image into separate channel images while preserving affine transforms."""
    if not stack.rgb:
        raise ValueError(
            trans._('Image must be RGB to use split_rgb', deferred=True)
        )

    data, meta, _ = stack.as_layer_data_tuple()

    meta['colormap'] = ('red', 'green', 'blue', 'gray')
    meta['rgb'] = False

    layerdata_list = split_channels(data, channel_axis=-1, **meta)

    images = [
        Image(image, **i_kwargs) for image, i_kwargs, _ in layerdata_list
    ]
    return images if with_alpha else images[:3]


def images_to_stack(images: list[Image], axis: int = 0, **kwargs) -> Image:
    """Combines a list of Image layers into one layer stacked along axis

    The new image layer will get the meta properties of the first
    image layer in the input list unless specified in kwargs

    Parameters
    ----------
    images : List
        List of Image Layers
    axis : int
        Index to to insert the new axis
    **kwargs : dict
        Dictionary of parameters values to override parameters
        from the first image in images list.

    Returns
    -------
    stack : napari.layers.Image
        Combined image stack
    """

    if not images:
        raise IndexError(trans._('images list is empty', deferred=True))

    if not all(isinstance(layer, Image) for layer in images):
        non_image_layers = [
            (layer.name, type(layer).__name__)
            for layer in images
            if not isinstance(layer, Image)
        ]
        raise ValueError(
            trans._(
                'All selected layers to be merged must be Image layers. '
                'The following layers are not Image layers: '
                f'{", ".join(f"{name} ({layer_type})" for name, layer_type in non_image_layers)}'
            )
        )

    data, meta, _ = images[0].as_layer_data_tuple()

    # RGB images do not need extra dimensions inserted into metadata
    if 'rgb' not in kwargs:
        kwargs.setdefault('scale', np.insert(meta['scale'], axis, 1))
        kwargs.setdefault('translate', np.insert(meta['translate'], axis, 0))

    meta.update(kwargs)

    # Check if input images are either all multiscale or not
    multiscale_flags = [
        getattr(image, 'multiscale', False) for image in images
    ]
    if not all(multiscale_flags) and any(multiscale_flags):
        raise ValueError(
            trans._(
                'All images must have the same multiscale status (all True or all False) to be stacked.\nGot: {multiscale_flags}',
                multiscale_flags=multiscale_flags[::-1],
                deferred=True,
            )
        )

    if all(multiscale_flags):
        # Check that all multiscale images have the same number of levels
        n_scales_list = [len(image.data) for image in images]
        if len(set(n_scales_list)) != 1:
            raise ValueError(
                trans._(
                    'All multiscale images must have the same number of levels to be stacked.\nGot: {n_scales_list}',
                    deferred=True,
                    n_scales_list=n_scales_list,
                )
            )
        arrays_to_check = [img.data[0] for img in images]
    else:
        arrays_to_check = [img.data for img in images]

    # check if any of the data arrays are zarr arrays
    # zarr doesn't have a stack method, so we will use dask.array.stack
    is_zarr = any(
        hasattr(arr, '__module__') and arr.__module__.startswith('zarr')
        for arr in arrays_to_check
    )

    if is_zarr:
        import dask.array as da

        stacker = da.stack
        warnings.warn(
            trans._(
                'zarr array cannot be stacked lazily, using dask array to stack.',
                deferred=True,
            )
        )
    else:
        stacker = np.stack

    if all(multiscale_flags):
        n_scales = len(images[0].data)
        new_data = [
            stacker([image.data[level] for image in images], axis=axis)
            for level in range(n_scales)
        ]
    else:
        new_data = stacker([image.data for image in images], axis=axis)

    # RGB images do not need extra dimensions inserted into metadata
    # They can use the meta dict from one of the source image layers
    if not meta['rgb']:
        meta['units'] = (pint.get_application_registry().pixel,) + meta[
            'units'
        ]
        meta['axis_labels'] = (f'axis -{data.ndim + 1}',) + meta['axis_labels']

    return Image(new_data, **meta)


def merge_rgb(images: list[Image]) -> Image:
    """Variant of images_to_stack that makes an RGB from 3 images."""
    if not (
        len(images) in [3, 4] and all(isinstance(x, Image) for x in images)
    ):
        raise ValueError(
            trans._(
                'Merging to RGB requires either 3 or 4 Image layers',
                deferred=True,
            )
        )
    if not all(image.data.shape == images[0].data.shape for image in images):
        all_shapes = [(image.name, image.data.shape) for image in images]
        raise ValueError(
            trans._(
                'Shape mismatch! To merge to RGB, all selected Image layers (with R, G, and B colormaps) must have the same shape. '
                'Mismatched shapes: '
                f'{", ".join(f"{name} (shape: {shape})" for name, shape in all_shapes)}'
            )
        )

    # we will check for the presence of R G B colormaps to determine how to merge
    colormaps = {image.colormap.name for image in images}
    r_g_b = ['red', 'green', 'blue']
    # if image is rgba, add gray colormap to represent alpha channel
    if len(colormaps) == 4:
        r_g_b.append('gray')
    if colormaps != set(r_g_b):
        missing_colormaps = set(r_g_b) - colormaps
        raise ValueError(
            trans._(
                'Missing colormap(s): {missing_colormaps}! To merge layers to '
                f'{"RGB" if len(r_g_b) == 3 else "RGBA"}, ensure you have '
                f'{", ".join(r_g_b[:-1])}, and {r_g_b[-1]} as layer colormaps.',
                missing_colormaps=missing_colormaps,
                deferred=True,
            )
        )

    # use the R G B colormaps to order the images for merging
    imgs = [
        image
        for color in r_g_b
        for image in images
        if image.colormap.name == color
    ]
    return images_to_stack(imgs, axis=-1, rgb=True)



from qtpy.QtWidgets import QCheckBox, QWidget

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import checked_to_bool, qt_signals_blocked
from napari.layers.base.base import Layer
from napari.utils.events import disconnect_events
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtTextVisibilityControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the text visibility
    layer attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Layer
        An instance of a napari layer.

    Attributes
    ----------
    text_disp_checkbox : qtpy.QtWidgets.QCheckbox
        Checkbox controlling if text on the layer is visible or not.
    text_disp_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the text visibility widget.
    """

    def __init__(self, parent: QWidget, layer: Layer) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.text.events.visible.connect(
            self._on_text_visibility_change
        )

        # Setup widgets
        text_disp_cb = QCheckBox()
        text_disp_cb.setToolTip(trans._('Toggle text visibility'))
        text_disp_cb.setChecked(self._layer.text.visible)
        connect_setattr(
            text_disp_cb.stateChanged,
            layer.text,
            'visible',
            convert_fun=checked_to_bool,
        )
        self.text_disp_checkbox = text_disp_cb
        self.text_disp_label = QtWrappedLabel(trans._('display text:'))

    def _on_text_visibility_change(self) -> None:
        """Receive layer model text visibiltiy change change event and update checkbox."""
        with qt_signals_blocked(self.text_disp_checkbox):
            self.text_disp_checkbox.setChecked(self._layer.text.visible)

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.text_disp_label, self.text_disp_checkbox)]

    def disconnect_widget_controls(self) -> None:
        disconnect_events(self._layer.text.events, self)
        super().disconnect_widget_controls()



from qtpy.QtWidgets import QComboBox, QWidget

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import qt_signals_blocked
from napari.layers import Points
from napari.layers.points._points_constants import (
    SYMBOL_TRANSLATION,
    SYMBOL_TRANSLATION_INVERTED,
)
from napari.utils.translations import trans


class QtSymbolComboBoxControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the current symbol
    layer attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Points
        An instance of a napari Points layer.

    Attributes
    ----------
    symbol_combobox : qtpy.QtWidgets.QComboBox
        Combobox controlling current symbol attribute of the layer.
    symbol_combobox_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the current symbol chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Points) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.symbol.connect(self._on_current_symbol_change)
        self._layer.events.current_symbol.connect(
            self._on_current_symbol_change
        )

        # Setup widgets
        sym_cb = QComboBox()
        sym_cb.setToolTip(
            trans._(
                'Change the symbol of currently selected points and any added afterwards.'
            )
        )
        current_index = 0
        for index, (symbol_string, text) in enumerate(
            SYMBOL_TRANSLATION.items()
        ):
            symbol_string = symbol_string.value
            sym_cb.addItem(text, symbol_string)

            if symbol_string == self._layer.current_symbol:
                current_index = index

        sym_cb.setCurrentIndex(current_index)
        sym_cb.currentTextChanged.connect(self.change_current_symbol)
        self.symbol_combobox = sym_cb

        self.symbol_combobox_label = QtWrappedLabel(trans._('symbol:'))

    def change_current_symbol(self, text: str) -> None:
        """Change marker symbol of the points on the layer model.

        Parameters
        ----------
        text : int
            Index of current marker symbol of points, eg: '+', '.', etc.
        """
        with self._layer.events.symbol.blocker(self._on_current_symbol_change):
            self._layer.current_symbol = SYMBOL_TRANSLATION_INVERTED[text]

    def _on_current_symbol_change(self) -> None:
        """Receive marker symbol change event and update the dropdown menu."""
        with qt_signals_blocked(self.symbol_combobox):
            self.symbol_combobox.setCurrentIndex(
                self.symbol_combobox.findData(self._layer.current_symbol.value)
            )

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.symbol_combobox_label, self.symbol_combobox)]



from qtpy.QtWidgets import (
    QComboBox,
    QWidget,
)

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import qt_signals_blocked
from napari.layers import Surface
from napari.layers.surface._surface_constants import SHADING_TRANSLATION
from napari.utils.translations import trans


class QtShadingComboBoxControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the layer shading
    value attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Surface
        An instance of a napari Surface layer.

    Attributes
    ----------
    shading_combobox : qtpy.QtWidgets.QComboBox
        ComboBox controlling current shading value of the layer.
    shading_combobox_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the shading value chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Surface) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.shading.connect(self._on_shading_change)

        # Setup widgets
        shading_comboBox = QComboBox(parent)
        for display_name, shading in SHADING_TRANSLATION.items():
            shading_comboBox.addItem(display_name, shading)
        index = shading_comboBox.findData(
            SHADING_TRANSLATION[self._layer.shading]
        )
        shading_comboBox.setCurrentIndex(index)
        shading_comboBox.currentTextChanged.connect(self.change_shading)
        self.shading_combobox = shading_comboBox

        self.shading_combobox_label = QtWrappedLabel(trans._('shading:'))

    def change_shading(self, text: str) -> None:
        """Change shading value on the surface layer.
        Parameters
        ----------
        text : str
            Name of shading mode, eg: 'flat', 'smooth', 'none'.
        """
        self._layer.shading = self.shading_combobox.currentData()

    def _on_shading_change(self) -> None:
        """Receive layer model shading change event and update combobox."""
        with qt_signals_blocked(self.shading_combobox):
            self.shading_combobox.setCurrentIndex(
                self.shading_combobox.findData(
                    SHADING_TRANSLATION[self._layer.shading]
                )
            )

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.shading_combobox_label, self.shading_combobox)]



from collections.abc import Iterable

import numpy as np
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QWidget
from superqt import QLabeledSlider

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import qt_signals_blocked
from napari.layers import Shapes
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtEdgeWidthSliderControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the current edge
    width layer attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Shapes
        An instance of a napari Shapes layer.

    Attributes
    ----------
    edge_width_slider : superqt.QLabeledDoubleSlider
        Slider controlling line edge width of layer.
    edge_width_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the current edge width widget.
    """

    def __init__(self, parent: QWidget, layer: Shapes) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.edge_width.connect(self._on_edge_width_change)

        # Setup widgets
        sld = QLabeledSlider(Qt.Orientation.Horizontal)
        sld.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        sld.setMinimum(0)
        sld.setMaximum(40)
        sld.setSingleStep(1)
        value = self._layer.current_edge_width
        if isinstance(value, Iterable):
            if isinstance(value, list):
                value = np.asarray(value)
            value = value.mean()
        sld.setValue(int(value))
        self.edge_width_slider = sld
        connect_setattr(
            self.edge_width_slider.valueChanged,
            self._layer,
            'current_edge_width',
            convert_fun=float,
        )
        self.edge_width_slider.setToolTip(
            trans._(
                'Set the edge width of currently selected shapes and any added afterwards.'
            )
        )
        self.edge_width_label = QtWrappedLabel(trans._('edge width:'))

    def _on_edge_width_change(self) -> None:
        """Receive layer model edge line width change event and update slider."""
        with qt_signals_blocked(self.edge_width_slider):
            value = self._layer.current_edge_width
            value = np.clip(int(value), 0, 40)
            self.edge_width_slider.setValue(value)

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.edge_width_label, self.edge_width_slider)]



from abc import ABC, abstractclassmethod

from qtpy.QtCore import QObject, Qt
from qtpy.QtWidgets import QLabel, QWidget

from napari.layers.base.base import Layer
from napari.utils.events import disconnect_events


class QtWrappedLabel(QLabel):
    """
    QLabel subclass with the `wordWrap` activated (True) and text aligned
    to the right and vertically centered by default.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.setWordWrap(True)
        self.setAlignment(
            Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter
        )


class MetaWidgetControlsBase(type(ABC), type(QObject)):
    pass


class QtWidgetControlsBase(QObject, ABC, metaclass=MetaWidgetControlsBase):
    """
    Base class that defines base methods for wrapper classes that do the
    connection of events/signals between layer attributes and Qt widgets.

    This class is never directly instantiated anywhere.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Layer
        An instance of a napari layer.
    """

    def __init__(self, parent: QWidget, layer: Layer) -> None:
        super().__init__(parent)
        # Setup layer
        self._layer = layer
        # Track registered callbacks (defined via `attr_to_settr` for example)
        # so it is possible to disconnect them when the widget is being closed/deleted
        self._callbacks = []

    @abstractclassmethod
    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        """
        Enable access to the created labels and control widgets.

        Returns
        -------
        list : list[tuple[QtWrappedLabel, QWidget]]
            List of tuples of the label and widget controls available.

        """

    def disconnect_widget_controls(self) -> None:
        """
        Disconnect layer from widget controls.
        """
        disconnect_events(self._layer.events, self)
        for callback in self._callbacks:
            disconnect_events(self._layer.events, callback)

    def deleteLater(self) -> None:
        self.disconnect_widget_controls()
        super().deleteLater()



import hashlib
import os
import shutil
import sys
from collections.abc import Callable
from functools import partial
from importlib.metadata import version
from pathlib import Path

import appdirs
from packaging.version import InvalidVersion, parse as parse_version

__all__ = (
    'user_cache_dir',
    'user_config_dir',
    'user_data_dir',
    'user_log_dir',
    'user_state_dir',
)

PREFIX_PATH = os.path.realpath(sys.prefix)

UV_POSSIBLE_PATH = appdirs.user_cache_dir('uv')

if PREFIX_PATH.startswith(UV_POSSIBLE_PATH):
    environment_marker = os.path.join(
        'uvx', parse_version(version('napari')).base_version
    )
else:
    environment_marker = f'{os.path.basename(PREFIX_PATH)}_{hashlib.sha1(PREFIX_PATH.encode()).hexdigest()}'

_appname = 'napari'
_appauthor = False


# all of these also take an optional "version" argument ... but if we want
# to be able to update napari while using data (e.g. plugins, settings) from
# an earlier version, we should leave off the version.

user_data_dir: Callable[[], str] = partial(
    appdirs.user_data_dir, _appname, _appauthor
)
user_config_dir: Callable[[], str] = partial(
    appdirs.user_config_dir, _appname, _appauthor, environment_marker
)
user_cache_dir: Callable[[], str] = partial(
    appdirs.user_cache_dir, _appname, _appauthor, environment_marker
)
user_state_dir: Callable[[], str] = partial(
    appdirs.user_state_dir, _appname, _appauthor
)
user_log_dir: Callable[[], str] = partial(
    appdirs.user_log_dir, _appname, _appauthor
)


def _maybe_migrate_uvx_settings() -> bool:
    """If we are in an uv environment, and there are no settings in the
    current config dir, but there are settings in the uvx config dir,
    move them over.

    Returns
    -------
    bool
        Whether or not a migration was performed.
    """
    if not environment_marker.startswith('uvx'):
        return False  # only migrate if we are in an uvx environment

    config_path = Path(user_config_dir())
    if config_path.exists():
        return False  # nothing to do, the current config path already exists

    base_config_path = config_path.parent

    if not base_config_path.exists():
        return False  # nothing to do, the base config path doesn't exist

    napari_version = parse_version(version('napari'))
    older_versions = []

    for fname in base_config_path.iterdir():
        try:
            dir_version = parse_version(fname.name)
        except InvalidVersion:
            continue
        if dir_version < napari_version:
            older_versions.append((dir_version, fname))

    if not older_versions:
        return False  # nothing to do, there are no older versions

    # get the latest version that is older than the current version
    shutil.copytree(max(older_versions)[1], config_path)
    return True



from qtpy.QtWidgets import QComboBox, QWidget

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import qt_signals_blocked
from napari.layers import Tracks
from napari.utils.colormaps import AVAILABLE_COLORMAPS
from napari.utils.translations import trans


class QtColormapComboBoxControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the layer colormaps
    attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Tracks
        An instance of a napari Tracks layer.

    Attributes
    ----------
    colormap_combobox : qtpy.QtWidgets.QComboBox
        ComboBox controlling current colormap of the layer.
    colormap_combobox_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the colormap chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Tracks) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.colormap.connect(self._on_colormap_change)

        # Setup widgets
        self.colormap_combobox = QComboBox()
        for name, colormap in AVAILABLE_COLORMAPS.items():
            display_name = colormap._display_name
            self.colormap_combobox.addItem(display_name, name)
        self.colormap_combobox.currentTextChanged.connect(self.change_colormap)

        self.colormap_combobox_label = QtWrappedLabel(trans._('colormap:'))

        self._on_colormap_change()

    def change_colormap(self, colormap: str):
        self._layer.colormap = self.colormap_combobox.currentData()

    def _on_colormap_change(self):
        """Receive layer model colormap change event and update combobox."""
        with qt_signals_blocked(self.colormap_combobox):
            self.colormap_combobox.setCurrentIndex(
                self.colormap_combobox.findData(self._layer.colormap)
            )

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.colormap_combobox_label, self.colormap_combobox)]



from qtpy.QtWidgets import (
    QCheckBox,
    QWidget,
)

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import attr_to_settr, checked_to_bool
from napari.layers import Labels
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtContiguousCheckBoxControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the layer contiguous
    attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Labels
        An instance of a napari Labels layer.

    Attributes
    ----------
    contiguous_checkbox : qtpy.QtWidgets.QCheckBox
        Checkbox to control if label layer is contiguous.
    contiguous_checkbox_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the contiguous model chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Labels) -> None:
        super().__init__(parent, layer)
        # Setup widgets
        contig_cb = QCheckBox()
        contig_cb.setToolTip(trans._('Contiguous editing'))
        self._callbacks.append(
            attr_to_settr(
                self._layer,
                'contiguous',
                contig_cb,
                'setChecked',
            )
        )
        connect_setattr(
            contig_cb.stateChanged,
            layer,
            'contiguous',
            convert_fun=checked_to_bool,
        )
        self.contiguous_checkbox = contig_cb

        self.contiguous_checkbox_label = QtWrappedLabel(trans._('contiguous:'))

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.contiguous_checkbox_label, self.contiguous_checkbox)]



from napari._qt.layer_controls.widgets._labels.qt_brush_size_slider import (
    QtBrushSizeSliderControl,
)
from napari._qt.layer_controls.widgets._labels.qt_color_mode_combobox import (
    QtColorModeComboBoxControl,
)
from napari._qt.layer_controls.widgets._labels.qt_contiguous_checkbox import (
    QtContiguousCheckBoxControl,
)
from napari._qt.layer_controls.widgets._labels.qt_contour_spinbox import (
    QtContourSpinBoxControl,
)
from napari._qt.layer_controls.widgets._labels.qt_display_selected_label_checkbox import (
    QtDisplaySelectedLabelCheckBoxControl,
)
from napari._qt.layer_controls.widgets._labels.qt_label_color import (
    QtLabelControl,
)
from napari._qt.layer_controls.widgets._labels.qt_ndim_spinbox import (
    QtNdimSpinBoxControl,
)
from napari._qt.layer_controls.widgets._labels.qt_preserve_labels_checkbox import (
    QtPreserveLabelsCheckBoxControl,
)
from napari._qt.layer_controls.widgets._labels.qt_render_control import (
    QtLabelRenderControl,
)

__all__ = [
    'QtBrushSizeSliderControl',
    'QtColorModeComboBoxControl',
    'QtContiguousCheckBoxControl',
    'QtContourSpinBoxControl',
    'QtDisplaySelectedLabelCheckBoxControl',
    'QtLabelControl',
    'QtLabelRenderControl',
    'QtNdimSpinBoxControl',
    'QtPreserveLabelsCheckBoxControl',
]



from dataclasses import dataclass, field
from unittest.mock import MagicMock

import numpy as np
import pytest

from napari.components.dims import Dims
from napari.layers import Points
from napari.layers._tests._utils import compare_dicts
from napari.layers.base import ActionType
from napari.utils._proxies import ReadOnlyWrapper
from napari.utils.interactions import (
    mouse_move_callbacks,
    mouse_press_callbacks,
    mouse_release_callbacks,
)


@dataclass
class Event:
    """Create a subclass for simulating vispy mouse events."""

    type: str
    is_dragging: bool = False
    modifiers: list[str] = field(default_factory=list)
    position: tuple[int, int] | tuple[int, int, int] = (
        0,
        0,
    )  # world coords
    pos: np.ndarray = field(
        default_factory=lambda: np.zeros(2)
    )  # canvas coords
    view_direction: list[float] | None = None
    up_direction: list[float] | None = None
    dims_displayed: list[int] = field(default_factory=lambda: [0, 1])


def read_only_event(*args, **kwargs):
    return ReadOnlyWrapper(Event(*args, **kwargs), exceptions=('handled',))


@pytest.fixture
def create_known_points_layer_2d():
    """Create points layer with known coordinates

    Returns
    -------
    layer : napari.layers.Points
        Points layer.
    n_points : int
        Number of points in the points layer
    known_non_point : list
        Data coordinates that are known to contain no points. Useful during
        testing when needing to guarantee no point is clicked on.
    """
    data = [[1, 3], [8, 4], [10, 10], [15, 4]]
    known_non_point = [10, 11]
    n_points = len(data)

    layer = Points(data, size=1)
    np.testing.assert_array_equal(layer.data, data)
    assert layer.ndim == 2
    assert len(layer.data) == n_points
    assert len(layer.selected_data) == 0

    return layer, n_points, known_non_point


@pytest.fixture
def create_known_points_layer_3d():
    """Create 3D points layer with known coordinates displayed in 3D.

    Returns
    -------
    layer : napari.layers.Points
        Points layer.
    n_points : int
        Number of points in the points layer
    known_non_point : list
        Data coordinates that are known to contain no points. Useful during
        testing when needing to guarantee no point is clicked on.
    """
    data = [[1, 2, 3], [8, 6, 4], [10, 5, 10], [15, 8, 4]]
    known_non_point = [4, 5, 6]
    n_points = len(data)

    layer = Points(data, size=1)
    layer._slice_dims(Dims(ndim=3, ndisplay=3))

    np.testing.assert_array_equal(layer.data, data)
    assert layer.ndim == 3
    assert len(layer._slice_input.displayed) == 3
    assert len(layer.data) == n_points
    assert len(layer._view_size) == n_points
    assert len(layer.selected_data) == 0

    return layer, n_points, known_non_point


def test_not_adding_or_selecting_point(create_known_points_layer_2d):
    """Don't add or select a point by clicking on one in pan_zoom mode."""
    layer, n_points, _ = create_known_points_layer_2d
    layer.mode = 'pan_zoom'

    # Simulate click
    event = read_only_event(type='mouse_press')
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_event(type='mouse_release')
    mouse_release_callbacks(layer, event)

    # Check no new point added and non selected
    assert len(layer.data) == n_points
    assert len(layer.selected_data) == 0


def test_add_point(create_known_points_layer_2d):
    """Add point by clicking in add mode."""
    layer, n_points, known_non_point = create_known_points_layer_2d

    # Add point at location where non exists
    layer.mode = 'add'

    # Simulate click
    event = read_only_event(type='mouse_press', position=known_non_point)
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_event(type='mouse_release', position=known_non_point)
    mouse_release_callbacks(layer, event)

    # Check new point added at coordinates location
    assert len(layer.data) == n_points + 1
    np.testing.assert_allclose(layer.data[-1], known_non_point)


def test_add_point_3d(create_known_points_layer_3d):
    """Add a point by clicking in 3D mode."""
    layer, n_points, known_not_point = create_known_points_layer_3d

    layer.mode = 'add'

    # Simulate click
    event = read_only_event(
        type='mouse_press',
        position=known_not_point,
        view_direction=[1, 0, 0],
        dims_displayed=[0, 1, 2],
    )
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_event(type='mouse_release', position=known_not_point)
    mouse_release_callbacks(layer, event)

    # Check clicked point selected
    assert len(layer.data) == (n_points + 1)
    np.testing.assert_array_equal(layer.data[-1], known_not_point)


def test_drag_in_add_mode(create_known_points_layer_2d):
    """Drag in add mode and make sure no point is added."""
    layer, n_points, known_non_point = create_known_points_layer_2d

    # Add point at location where non exists
    layer.mode = 'add'

    # Simulate click
    event = read_only_event(
        type='mouse_press', position=known_non_point, pos=np.array([0, 0])
    )
    mouse_press_callbacks(layer, event)

    known_non_point_end = [40, 60]

    # Simulate drag end
    event = read_only_event(
        type='mouse_move',
        is_dragging=True,
        position=known_non_point_end,
        pos=np.array([4, 4]),
    )
    mouse_move_callbacks(layer, event)

    # Simulate release
    event = read_only_event(
        type='mouse_release',
        position=known_non_point_end,
        pos=np.array([4, 4]),
    )
    mouse_release_callbacks(layer, event)

    # Check that no new point has been added
    assert len(layer.data) == n_points


def test_select_point(create_known_points_layer_2d):
    """Select a point by clicking on one in select mode."""
    layer = create_known_points_layer_2d[0]

    layer.mode = 'select'
    position = tuple(layer.data[0])

    # Simulate click
    event = read_only_event(type='mouse_press', position=position)
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_event(type='mouse_release', position=position)
    mouse_release_callbacks(layer, event)

    # Check clicked point selected
    assert len(layer.selected_data) == 1
    assert 0 in layer.selected_data


def test_select_point_3d(create_known_points_layer_3d):
    """Select a point by clicking on one in select mode in 3D mode."""
    layer = create_known_points_layer_3d[0]

    layer.mode = 'select'
    position = tuple(layer.data[1])

    # Simulate click
    event = read_only_event(
        type='mouse_press',
        position=position,
        view_direction=[1, 0, 0],
        dims_displayed=[0, 1, 2],
    )
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_event(type='mouse_release', position=position)
    mouse_release_callbacks(layer, event)

    # Check clicked point selected
    assert len(layer.selected_data) == 1
    assert 1 in layer.selected_data


def test_unselect_by_click_point_3d(create_known_points_layer_3d):
    """Select unselecting point by shift clicking on it again in 3D mode."""
    layer = create_known_points_layer_3d[0]

    layer.mode = 'select'
    position = tuple(layer.data[1])

    layer.selected_data = {0, 1}

    # Simulate shift+click on point 1
    event = read_only_event(
        type='mouse_press',
        position=position,
        modifiers=['Shift'],
        view_direction=[1, 0, 0],
        dims_displayed=[0, 1, 2],
    )
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_event(
        type='mouse_release', modifiers=['Shift'], position=position
    )
    mouse_release_callbacks(layer, event)

    # Check clicked point selected
    assert layer.selected_data == {0}


def test_select_by_shift_click_3d(create_known_points_layer_3d):
    """Select selecting point by shift clicking on an additional point in 3D"""
    layer = create_known_points_layer_3d[0]

    layer.mode = 'select'
    position = tuple(layer.data[1])

    layer.selected_data = {0}

    # Simulate shift+click on point 1
    event = read_only_event(
        type='mouse_press',
        position=position,
        modifiers=['Shift'],
        view_direction=[1, 0, 0],
        dims_displayed=[0, 1, 2],
    )
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_event(
        type='mouse_release', modifiers=['Shift'], position=position
    )
    mouse_release_callbacks(layer, event)

    # Check clicked point selected
    assert layer.selected_data == {0, 1}


def test_unselect_by_click_empty_3d(create_known_points_layer_3d):
    """Select unselecting point by clicking in empty space"""
    layer, _n_points, known_not_point = create_known_points_layer_3d

    layer.mode = 'select'

    layer.selected_data = {0, 1}

    # Simulate click on point
    event = read_only_event(
        type='mouse_press',
        position=known_not_point,
        view_direction=[1, 0, 0],
        dims_displayed=[0, 1, 2],
    )
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_event(type='mouse_release', position=known_not_point)
    mouse_release_callbacks(layer, event)

    # Check clicked point selected
    assert len(layer.selected_data) == 0


def test_after_in_add_mode_point(create_known_points_layer_2d):
    """Don't add or select a point by clicking on one in pan_zoom mode."""
    layer, n_points, _ = create_known_points_layer_2d

    layer.mode = 'add'
    layer.mode = 'pan_zoom'
    position = tuple(layer.data[0])

    # Simulate click
    event = read_only_event(type='mouse_press', position=position)
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_event(type='mouse_release', position=position)
    mouse_release_callbacks(layer, event)

    # Check no new point added and non selected
    assert len(layer.data) == n_points
    assert len(layer.selected_data) == 0


def test_after_in_select_mode_point(create_known_points_layer_2d):
    """Don't add or select a point by clicking on one in pan_zoom mode."""
    layer, n_points, _ = create_known_points_layer_2d

    layer.mode = 'select'
    layer.mode = 'pan_zoom'
    position = tuple(layer.data[0])

    # Simulate click
    event = read_only_event(type='mouse_press', position=position)
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_event(type='mouse_release', position=position)
    mouse_release_callbacks(layer, event)

    # Check no new point added and non selected
    assert len(layer.data) == n_points
    assert len(layer.selected_data) == 0


def test_unselect_select_point(create_known_points_layer_2d):
    """Select a point by clicking on one in select mode."""
    layer = create_known_points_layer_2d[0]

    layer.mode = 'select'
    position = tuple(layer.data[0])
    layer.selected_data = {2, 3}

    # Simulate click
    event = read_only_event(type='mouse_press', position=position)
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_event(type='mouse_release', position=position)
    mouse_release_callbacks(layer, event)

    # Check clicked point selected
    assert len(layer.selected_data) == 1
    assert 0 in layer.selected_data


def test_add_select_point(create_known_points_layer_2d):
    """Add to a selection of points point by shift-clicking on one."""
    layer = create_known_points_layer_2d[0]

    layer.mode = 'select'
    position = tuple(layer.data[0])
    layer.selected_data = {2, 3}

    # Simulate click
    event = read_only_event(
        type='mouse_press', modifiers=['Shift'], position=position
    )
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_event(
        type='mouse_release', modifiers=['Shift'], position=position
    )
    mouse_release_callbacks(layer, event)

    # Check clicked point selected
    assert len(layer.selected_data) == 3
    assert layer.selected_data == {2, 3, 0}


def test_remove_select_point(create_known_points_layer_2d):
    """Remove from a selection of points point by shift-clicking on one."""
    layer = create_known_points_layer_2d[0]

    layer.mode = 'select'
    position = tuple(layer.data[0])
    layer.selected_data = {0, 2, 3}

    # Simulate click
    event = read_only_event(
        type='mouse_press', modifiers=['Shift'], position=position
    )
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_event(
        type='mouse_release', modifiers=['Shift'], position=position
    )
    mouse_release_callbacks(layer, event)

    # Check clicked point selected
    assert len(layer.selected_data) == 2
    assert layer.selected_data == {2, 3}


def test_not_selecting_point(create_known_points_layer_2d):
    """Don't select a point by not clicking on one in select mode."""
    layer, _n_points, known_non_point = create_known_points_layer_2d

    layer.mode = 'select'

    # Simulate click
    event = read_only_event(type='mouse_press', position=known_non_point)
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_event(type='mouse_release', position=known_non_point)
    mouse_release_callbacks(layer, event)

    # Check clicked point selected
    assert len(layer.selected_data) == 0


def test_unselecting_points(create_known_points_layer_2d):
    """Unselect points by not clicking on one in select mode."""
    layer, _n_points, known_non_point = create_known_points_layer_2d

    layer.mode = 'select'
    layer.selected_data = {2, 3}
    assert len(layer.selected_data) == 2

    # Simulate click
    event = read_only_event(type='mouse_press', position=known_non_point)
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_event(type='mouse_release', position=known_non_point)
    mouse_release_callbacks(layer, event)

    # Check clicked point selected
    assert len(layer.selected_data) == 0

    # check that this also works with scaled data and position near a point (see #5737)
    # we are taking the first point and shifting *slightly* more than the point size
    layer.scale = 100, 100
    pos = np.array(layer.data[0])
    pos[1] += layer.size[0] * 2

    event = read_only_event(type='mouse_press', position=pos)
    mouse_press_callbacks(layer, event)
    event = read_only_event(type='mouse_release', position=pos)
    mouse_release_callbacks(layer, event)

    # Check clicked point selected
    assert len(layer.selected_data) == 0


def test_selecting_all_points_with_drag_2d(create_known_points_layer_2d):
    """Select all points when drag box includes all of them."""
    layer, n_points, _known_non_point = create_known_points_layer_2d

    layer.mode = 'select'

    # drag a box that includes all the points
    box_drag_begin = (20, 20)
    box_drag_end = (0, 0)

    # Simulate click
    event = read_only_event(type='mouse_press', position=box_drag_begin)
    mouse_press_callbacks(layer, event)

    # Simulate drag start
    event = read_only_event(
        type='mouse_move', is_dragging=True, position=box_drag_begin
    )
    mouse_move_callbacks(layer, event)

    # Simulate drag end
    event = read_only_event(
        type='mouse_move', is_dragging=True, position=box_drag_end
    )
    mouse_move_callbacks(layer, event)

    # Simulate release
    event = read_only_event(
        type='mouse_release', is_dragging=True, position=box_drag_end
    )
    mouse_release_callbacks(layer, event)

    # Check all points selected as drag box contains them
    assert len(layer.selected_data) == n_points


def test_selecting_no_points_with_drag_2d(create_known_points_layer_2d):
    """Select no points when drag box outside of all of them."""
    layer, _n_points, known_non_point = create_known_points_layer_2d

    layer.mode = 'select'

    # Simulate click
    event = read_only_event(type='mouse_press', position=known_non_point)
    mouse_press_callbacks(layer, event)

    # Simulate drag start
    event = read_only_event(
        type='mouse_move', is_dragging=True, position=known_non_point
    )
    mouse_move_callbacks(layer, event)

    # Simulate drag end
    event = read_only_event(
        type='mouse_move', is_dragging=True, position=(50, 60)
    )
    mouse_move_callbacks(layer, event)

    # Simulate release
    event = read_only_event(
        type='mouse_release', is_dragging=True, position=(50, 60)
    )
    mouse_release_callbacks(layer, event)

    # Check no points selected as drag box doesn't contain them
    assert len(layer.selected_data) == 0


def test_selecting_points_with_drag_3d(create_known_points_layer_3d):
    """Select all points when drag box includes all of them."""
    layer = create_known_points_layer_3d[0]

    layer.mode = 'select'

    # Simulate click
    event = read_only_event(
        type='mouse_press',
        position=(5, 0, 0),
        view_direction=[1, 0, 0],
        up_direction=[0, 1, 0],
        dims_displayed=[0, 1, 2],
    )
    mouse_press_callbacks(layer, event)

    # Simulate drag start
    event = read_only_event(
        type='mouse_move',
        is_dragging=True,
        position=(5, 0, 0),
        view_direction=[1, 0, 0],
        up_direction=[0, 1, 0],
        dims_displayed=[0, 1, 2],
    )
    mouse_move_callbacks(layer, event)

    # Simulate drag end
    event = read_only_event(
        type='mouse_move',
        is_dragging=True,
        position=(5, 6, 6),
        view_direction=[1, 0, 0],
        up_direction=[0, 1, 0],
        dims_displayed=[0, 1, 2],
    )
    mouse_move_callbacks(layer, event)

    # Simulate release
    event = read_only_event(
        type='mouse_release',
        is_dragging=True,
        position=(5, 6, 6),
        view_direction=[1, 0, 0],
        up_direction=[0, 1, 0],
        dims_displayed=[0, 1, 2],
    )
    mouse_release_callbacks(layer, event)

    # Check all points selected as drag box contains them
    assert layer.selected_data == {0, 1}


def test_selecting_no_points_with_drag_3d(create_known_points_layer_3d):
    """Select no points when drag box outside of all of them."""
    layer = create_known_points_layer_3d[0]

    layer.mode = 'select'

    # Simulate click
    event = read_only_event(
        type='mouse_press',
        position=(5, 15, 15),
        view_direction=[1, 0, 0],
        up_direction=[0, 1, 0],
        dims_displayed=[0, 1, 2],
    )
    mouse_press_callbacks(layer, event)

    # Simulate drag start
    event = read_only_event(
        type='mouse_move',
        is_dragging=True,
        position=(5, 15, 15),
        view_direction=[1, 0, 0],
        up_direction=[0, 1, 0],
        dims_displayed=[0, 1, 2],
    )
    mouse_move_callbacks(layer, event)

    # Simulate drag end
    event = read_only_event(
        type='mouse_move',
        is_dragging=True,
        position=(5, 20, 20),
        view_direction=[1, 0, 0],
        up_direction=[0, 1, 0],
        dims_displayed=[0, 1, 2],
    )
    mouse_move_callbacks(layer, event)

    # Simulate release
    event = read_only_event(
        type='mouse_release',
        is_dragging=True,
        position=(5, 20, 20),
        view_direction=[1, 0, 0],
        up_direction=[0, 1, 0],
        dims_displayed=[0, 1, 2],
    )
    mouse_release_callbacks(layer, event)

    # Check all points selected as drag box contains them
    assert len(layer.selected_data) == 0


@pytest.mark.parametrize(
    ('pre_selection', 'on_point', 'modifier'),
    [
        (set(), True, []),
        ({0}, True, []),
        ({0, 1, 2}, True, []),
        ({1, 2}, True, []),
        (set(), True, ['Shift']),
        ({0}, True, ['Shift']),
        ({0, 1, 2}, True, ['Shift']),
        ({1, 2}, True, ['Shift']),
        (set(), False, []),
        ({0}, False, []),
        ({0, 1, 2}, False, []),
        ({1, 2}, False, []),
        (set(), False, ['Shift']),
        ({0}, False, ['Shift']),
        ({0, 1, 2}, False, ['Shift']),
        ({1, 2}, False, ['Shift']),
    ],
)
def test_drag_start_selection(
    create_known_points_layer_2d, pre_selection, on_point, modifier
):
    """Check layer drag start and drag box behave as expected."""
    layer, n_points, _known_non_point = create_known_points_layer_2d
    layer.mode = 'select'
    layer.selected_data = pre_selection

    initial_position = tuple(layer.data[0]) if on_point else (20, 20)
    zero_pos = [0, 0]
    initial_position_1 = tuple(layer.data[1])
    diff_data_1 = [
        layer.data[1, 0] - layer.data[0, 0],
        layer.data[1, 1] - layer.data[0, 1],
    ]

    assert layer._drag_start is None
    assert layer._drag_box is None

    # Simulate click
    event = read_only_event(
        type='mouse_press', position=initial_position, modifiers=modifier
    )
    mouse_press_callbacks(layer, event)

    if modifier:
        if not on_point:
            assert layer.selected_data == pre_selection
        elif 0 in pre_selection:
            assert layer.selected_data == pre_selection - {0}
        else:
            assert layer.selected_data == pre_selection | {0}
    elif not on_point:
        assert layer.selected_data == set()
    elif 0 in pre_selection:
        assert layer.selected_data == pre_selection
    else:
        assert layer.selected_data == {0}

    if len(layer.selected_data) > 0:
        center = layer.data[list(layer.selected_data), :].mean(axis=0)
    else:
        center = [0, 0]

    if not modifier:
        start_position = [
            initial_position[0] - center[0],
            initial_position[1] - center[1],
        ]
    else:
        start_position = initial_position

    is_point_move = len(layer.selected_data) > 0 and on_point and not modifier

    np.testing.assert_array_equal(layer._drag_start, start_position)

    # Simulate drag start on a different position
    offset_position = [initial_position[0] + 20, initial_position[1] + 20]
    event = read_only_event(
        type='mouse_move',
        is_dragging=True,
        position=offset_position,
        modifiers=modifier,
    )
    mouse_move_callbacks(layer, event)

    # Initial mouse_move is already considered a move and not a press.
    # Therefore, the _drag_start value should be identical and the data or drag_box should reflect
    # the mouse position.
    np.testing.assert_array_equal(layer._drag_start, start_position)
    if is_point_move:
        if 1 in layer.selected_data and 0 in layer.selected_data:
            np.testing.assert_array_equal(
                layer.data[1],
                [
                    offset_position[0] + diff_data_1[0],
                    offset_position[1] + diff_data_1[1],
                ],
            )
        elif 1 not in layer.selected_data:
            np.testing.assert_array_equal(layer.data[1], initial_position_1)

        if 0 in layer.selected_data:
            np.testing.assert_array_equal(
                layer.data[0], [offset_position[0], offset_position[1]]
            )
        else:
            raise AssertionError('Unreachable code')  # pragma: no cover
    else:
        np.testing.assert_array_equal(
            layer._drag_box, [initial_position, offset_position]
        )

    # Simulate drag start on new different position
    offset_position = zero_pos
    event = read_only_event(
        type='mouse_move',
        is_dragging=True,
        position=offset_position,
        modifiers=modifier,
    )
    mouse_move_callbacks(layer, event)

    # Initial mouse_move is already considered a move and not a press.
    # Therefore, the _drag_start value should be identical and the data or drag_box should reflect
    # the mouse position.
    np.testing.assert_array_equal(layer._drag_start, start_position)
    if is_point_move:
        if 1 in layer.selected_data and 0 in layer.selected_data:
            np.testing.assert_array_equal(
                layer.data[1],
                [
                    offset_position[0] + diff_data_1[0],
                    offset_position[1] + diff_data_1[1],
                ],
            )
        elif 1 not in layer.selected_data:
            np.testing.assert_array_equal(layer.data[1], initial_position_1)

        if 0 in layer.selected_data:
            np.testing.assert_array_equal(
                layer.data[0], [offset_position[0], offset_position[1]]
            )
        else:
            raise AssertionError('Unreachable code')  # pragma: no cover
    else:
        np.testing.assert_array_equal(
            layer._drag_box, [initial_position, offset_position]
        )

    # Simulate release
    event = read_only_event(
        type='mouse_release', is_dragging=True, modifiers=modifier
    )
    mouse_release_callbacks(layer, event)

    if on_point and 0 in pre_selection and modifier:
        assert layer.selected_data == pre_selection - {0}
    elif on_point and 0 in pre_selection and not modifier:
        assert layer.selected_data == pre_selection
    elif on_point and 0 not in pre_selection and modifier:
        assert layer.selected_data == pre_selection | {0}
    elif on_point and 0 not in pre_selection and not modifier:
        assert layer.selected_data == {0}
    elif 0 in pre_selection and modifier:
        assert 0 not in layer.selected_data
        assert layer.selected_data == (set(range(n_points)) - pre_selection)
    elif 0 in pre_selection and not modifier:
        assert 0 in layer.selected_data
        assert layer.selected_data == set(range(n_points))
    elif 0 not in pre_selection and modifier:
        assert 0 in layer.selected_data
        assert layer.selected_data == (set(range(n_points)) - pre_selection)
    elif 0 not in pre_selection and not modifier:
        assert 0 in layer.selected_data
        assert layer.selected_data == set(range(n_points))
    else:
        pytest.fail('Unreachable code')
    assert layer._drag_box is None
    assert layer._drag_start is None


def test_drag_point_with_mouse(create_known_points_layer_2d):
    layer = create_known_points_layer_2d[0]
    layer.events.data = MagicMock()
    layer.mode = 'select'
    old_data = (
        layer.data.copy()
    )  # ensure you have old data, not updated in place
    layer.selected_data = {1}
    initial_position = tuple(layer.data[1])

    new_position = [0, 0]
    modifier = []

    event = read_only_event(
        type='mouse_press', position=initial_position, modifiers=modifier
    )
    mouse_press_callbacks(layer, event)

    # Required to assert before the changing event as otherwise layer.data for changing is updated in place.
    changing_event = {
        'value': old_data,
        'action': ActionType.CHANGING,
        'data_indices': (1,),
        'vertex_indices': ((),),
    }

    def side_effect(*args, **kwargs):
        if kwargs['action'] == ActionType.CHANGING:
            assert compare_dicts(kwargs, changing_event)

    layer.events.data.side_effect = side_effect
    event = read_only_event(
        type='mouse_move',
        is_dragging=True,
        position=new_position,
        modifiers=modifier,
    )
    mouse_move_callbacks(layer, event)

    event = read_only_event(
        type='mouse_release', is_dragging=False, modifiers=modifier
    )
    mouse_release_callbacks(layer, event)

    changed_event = {
        'value': layer.data,
        'action': ActionType.CHANGED,
        'data_indices': (1,),
        'vertex_indices': ((),),
    }
    assert not np.array_equal(layer.data, old_data)
    assert compare_dicts(layer.events.data.call_args[1], changed_event)



from napari._vispy.overlays.zoom import VispyZoomOverlay
from napari.components.overlays import ZoomOverlay
from napari.components.viewer_model import ViewerModel


def test_zoom_overlay_initialization():
    viewer = ViewerModel()
    zoom_model = ZoomOverlay()
    zoom_view = VispyZoomOverlay(viewer=viewer, overlay=zoom_model)

    # change the visibility
    assert zoom_view.node is not None
    assert zoom_view.node.visible is False
    zoom_model.visible = True
    assert zoom_view.node.visible is True



from __future__ import annotations

import gc
import weakref
from collections.abc import Callable
from typing import TYPE_CHECKING

import numpy as np

from napari._tests.utils import skip_local_popups, skip_on_win_ci

if TYPE_CHECKING:
    from pytestqt.qtbot import QtBot

    from napari import Viewer

    MakeNapariViewer = Callable[..., Viewer]


@skip_local_popups
def test_memory_leaking(
    qtbot: QtBot, make_napari_viewer: MakeNapariViewer
) -> None:
    data = np.zeros((5, 20, 20, 20), dtype=int)
    data[1, 0:10, 0:10, 0:10] = 1
    viewer = make_napari_viewer()
    image = weakref.ref(viewer.add_image(data))
    labels = weakref.ref(viewer.add_labels(data))
    del viewer.layers[0]
    del viewer.layers[0]
    qtbot.wait(100)
    gc.collect()
    gc.collect()
    assert len(viewer.layers) == 0
    assert image() is None
    assert labels() is None


@skip_on_win_ci
@skip_local_popups
def test_leaks_image(
    qtbot: QtBot, make_napari_viewer: MakeNapariViewer
) -> None:
    viewer = make_napari_viewer(show=True)
    lr = weakref.ref(viewer.add_image(np.zeros((10, 10))))
    dr = weakref.ref(lr().data)

    viewer.layers.clear()
    qtbot.wait(100)
    gc.collect()
    gc.collect()
    assert lr() is None
    assert dr() is None


@skip_on_win_ci
@skip_local_popups
def test_leaks_labels(
    qtbot: QtBot, make_napari_viewer: MakeNapariViewer
) -> None:
    viewer = make_napari_viewer(show=True)
    rng = np.random.default_rng(0)
    lr = weakref.ref(
        viewer.add_labels(rng.integers(0, 10, size=(10, 10), dtype=np.int8)),
    )
    dr = weakref.ref(lr().data)
    viewer.layers.clear()
    qtbot.wait(100)
    gc.collect()
    gc.collect()
    assert lr() is None
    assert dr() is None



from typing import Optional

from qtpy.QtWidgets import QWidget

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import attr_to_settr
from napari._qt.widgets.qt_color_swatch import QColorSwatchEdit
from napari.layers import Shapes
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtEdgeColorControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the current edge
    color layer attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Shapes
        An instance of a napari Shapes layer.
    toolip : str
        String to use for the tooltip of the edge color edit widget.

    Attributes
    ----------
    edge_color_edit : napari._qt.widgets.qt_color_swatch.QColorSwatchEdit
        ColorSwatchEdit controlling current edge color of the layer.
    edge_color_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the current edge color chooser widget.
    """

    def __init__(
        self, parent: QWidget, layer: Shapes, tooltip: Optional[str] = None
    ) -> None:
        super().__init__(parent, layer)
        # Setup widgets
        self.edge_color_edit = QColorSwatchEdit(
            initial_color=self._layer.current_edge_color,
            tooltip=tooltip,
        )
        connect_setattr(
            self.edge_color_edit.color_changed,
            self._layer,
            'current_edge_color',
        )
        self._callbacks.append(
            attr_to_settr(
                self._layer,
                'current_edge_color',
                self.edge_color_edit,
                'setColor',
            )
        )
        self.edge_color_label = QtWrappedLabel(trans._('edge color:'))

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.edge_color_label, self.edge_color_edit)]



from __future__ import annotations

from typing import TYPE_CHECKING

from napari._qt.layer_controls.qt_layer_controls_base import QtLayerControls
from napari._qt.layer_controls.widgets import (
    QtColormapControl,
    QtContrastLimitsControl,
    QtGammaSliderControl,
)

if TYPE_CHECKING:
    from napari.layers import Image


class QtBaseImageControls(QtLayerControls):
    """Superclass for classes requiring colormaps, contrast & gamma sliders.

    This class is never directly instantiated anywhere.
    It is subclassed by QtImageControls and QtSurfaceControls.

    Parameters
    ----------
    layer : napari.layers.Image | napari.layers.Surface
        An instance of a napari Image or Surface layer.

    Attributes
    ----------
    _colormap_control : napari._qt.layer_controls.widgets.QtColormapControl
        Widget that wraps combobox and label widgets related with the layer colormap attribute.
    _contrast_limits_control : napari._qt.layer_controls.widgets.QtContrastLimitsControl
        Widget that wraps layer contrast range slider and autocontrast widgets.
    _gamma_slider_control : napari._qt.layer_controls.widgets.QtGammaSliderControl
        Widget that wraps layer gamma adjustment slider widget.
    """

    def __init__(self, layer: Image) -> None:
        super().__init__(layer)
        # Setup widgets controls
        self._contrast_limits_control = QtContrastLimitsControl(self, layer)
        self._add_widget_controls(self._contrast_limits_control)
        self._gamma_slider_control = QtGammaSliderControl(self, layer)
        self._add_widget_controls(self._gamma_slider_control)
        self._colormap_control = QtColormapControl(self, layer)
        self._add_widget_controls(self._colormap_control)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from vispy.scene import Axis, Image, Line, Node, STTransform

from napari._vispy.visuals.text import Text

if TYPE_CHECKING:
    from typing import Any

    import numpy.typing as npt
    from vispy.color import Colormap as VispyColormap

    from napari.utils.color import ColorValue


class Colormap(Node):
    _box_data = np.array(
        [
            [0, 0],
            [0, 1],
            [1, 1],
            [1, 0],
            [0, 0],
        ]
    )

    def __init__(self) -> None:
        super().__init__()
        self.ticks = Axis(
            pos=self._box_data[2:4],
            tick_direction=(1, 0),
            tick_width=2,
            tick_label_margin=4,
            axis_width=0,
            parent=self,
        )
        # override to use our class which works better with hidpi
        self.ticks.remove_subvisual(self.ticks._text)
        self.ticks._text = Text(
            font_size=self.ticks.tick_font_size, color=self.ticks.text_color
        )
        self.ticks.add_subvisual(self.ticks._text)

        self.ticks.transform = STTransform()
        self.img = Image(parent=self)
        self.img.transform = STTransform()
        self.box = Line(
            pos=self._box_data,
            connect='strip',
            method='gl',
            width=1,
            parent=self,
        )
        self.box.transform = STTransform()

        self._texture_size = 250
        self._text_vertices_size = (0, 0)
        self.set_data_and_clim()

    def set_data_and_clim(
        self,
        clim: tuple[float, float] = (0, 1),
        dtype: npt.DTypeLike = np.float32,
    ) -> None:
        self.img.set_data(
            np.linspace(
                clim[1], clim[0], self._texture_size, dtype=dtype
            ).reshape(-1, 1)
        )
        self.img.clim = clim

    def set_cmap(self, cmap: VispyColormap) -> None:
        self.img.cmap = cmap

    def set_gamma(self, gamma: float) -> None:
        self.img.gamma = gamma

    def set_gl_state(self, *args: Any, **kwargs: Any) -> None:
        self.img.set_gl_state(*args, **kwargs)

    def set_size(self, size: tuple[float, float]) -> None:
        self.img.transform.scale = size[0], size[1] / self._texture_size
        self.box.transform.scale = size
        self.ticks.transform.scale = size

    def set_ticks_and_get_text_size(
        self,
        tick_length: float,
        font_size: int,
        clim: tuple[float, float],
        color: ColorValue,
    ) -> tuple[float, float]:
        self.box.set_data(color=color)
        self.ticks.axis_color = color
        self.ticks.text_color = color
        self.ticks.tick_color = color
        self.ticks.tick_font_size = font_size
        self.ticks.major_tick_length = tick_length
        self.ticks.minor_tick_length = tick_length / 2
        self.ticks.domain = clim

        text = self.ticks._text
        self.ticks._update_subvisuals()  # triggers computing of the tick labels

        width, height = text.get_width_height()
        return width + self.ticks.tick_label_margin, height



from napari._qt.layer_controls.widgets._points.qt_border_color import (
    QtBorderColorControl,
)
from napari._qt.layer_controls.widgets._points.qt_current_size_slider import (
    QtCurrentSizeSliderControl,
)
from napari._qt.layer_controls.widgets._points.qt_symbol_combobox import (
    QtSymbolComboBoxControl,
)

__all__ = [
    'QtBorderColorControl',
    'QtCurrentSizeSliderControl',
    'QtSymbolComboBoxControl',
]



import numpy as np

from napari._qt.layer_controls.qt_image_controls import QtImageControls
from napari.components.dims import Dims
from napari.layers import Image


def test_interpolation_combobox(qtbot):
    """Changing the model attribute should update the view"""
    layer = Image(np.random.rand(8, 8))
    qtctrl = QtImageControls(layer)
    qtbot.addWidget(qtctrl)
    combo = qtctrl._interpolation_control.interpolation_combobox
    opts = {combo.itemText(i) for i in range(combo.count())}
    assert opts == {'cubic', 'linear', 'kaiser', 'nearest', 'spline36'}
    # programmatically adding approved interpolation works
    layer.interpolation2d = 'lanczos'
    assert combo.findText('lanczos') == 5


def test_rendering_combobox(qtbot):
    """Changing the model attribute should update the view"""
    layer = Image(np.random.rand(8, 8))
    qtctrl = QtImageControls(layer)
    qtbot.addWidget(qtctrl)
    combo = qtctrl._render_control.render_combobox
    opts = {combo.itemText(i) for i in range(combo.count())}
    rendering_options = {
        'translucent',
        'additive',
        'iso',
        'mip',
        'minip',
        'attenuated_mip',
        'average',
    }
    assert opts == rendering_options
    # programmatically updating rendering mode updates the combobox
    layer.rendering = 'iso'
    assert combo.findText('iso') == combo.currentIndex()


def test_depiction_combobox_changes(qtbot):
    """Changing the model attribute should update the view."""
    layer = Image(np.random.rand(10, 15, 20))
    qtctrl = QtImageControls(layer)
    qtctrl.ndisplay = 3
    qtbot.addWidget(qtctrl)
    combo_box = qtctrl._depiction_control.depiction_combobox
    opts = {combo_box.itemText(i) for i in range(combo_box.count())}
    depiction_options = {
        'volume',
        'plane',
    }
    assert opts == depiction_options
    layer.depiction = 'plane'
    assert combo_box.findText('plane') == combo_box.currentIndex()
    layer.depiction = 'volume'
    assert combo_box.findText('volume') == combo_box.currentIndex()


def test_plane_controls_show_hide_on_depiction_change(qtbot):
    """Changing depiction mode should show/hide plane controls in 3D."""
    layer = Image(np.random.rand(10, 15, 20))
    qtctrl = QtImageControls(layer)
    qtbot.addWidget(qtctrl)
    qtctrl.ndisplay = 3

    layer.depiction = 'volume'
    assert qtctrl._depiction_control.plane_thickness_slider.isHidden()
    assert qtctrl._depiction_control.plane_thickness_label.isHidden()
    assert qtctrl._depiction_control.plane_normal_buttons.isHidden()
    assert qtctrl._depiction_control.plane_normal_label.isHidden()

    layer.depiction = 'plane'
    assert not qtctrl._depiction_control.plane_thickness_slider.isHidden()
    assert not qtctrl._depiction_control.plane_thickness_label.isHidden()
    assert not qtctrl._depiction_control.plane_normal_buttons.isHidden()
    assert not qtctrl._depiction_control.plane_normal_label.isHidden()


def test_plane_controls_show_hide_on_ndisplay_change(qtbot):
    """Changing ndisplay should show/hide plane controls if depicting a plane."""
    layer = Image(np.random.rand(10, 15, 20))
    layer.depiction = 'plane'
    qtctrl = QtImageControls(layer)
    qtbot.addWidget(qtctrl)

    assert qtctrl.ndisplay == 2
    assert qtctrl._depiction_control.plane_thickness_slider.isHidden()
    assert qtctrl._depiction_control.plane_thickness_label.isHidden()
    assert qtctrl._depiction_control.plane_normal_buttons.isHidden()
    assert qtctrl._depiction_control.plane_normal_label.isHidden()

    qtctrl.ndisplay = 3
    assert not qtctrl._depiction_control.plane_thickness_slider.isHidden()
    assert not qtctrl._depiction_control.plane_thickness_label.isHidden()
    assert not qtctrl._depiction_control.plane_normal_buttons.isHidden()
    assert not qtctrl._depiction_control.plane_normal_label.isHidden()


def test_plane_slider_value_change(qtbot):
    """Changing the model should update the view."""
    layer = Image(np.random.rand(10, 15, 20))
    qtctrl = QtImageControls(layer)
    qtbot.addWidget(qtctrl)
    layer.plane.thickness *= 2
    assert (
        qtctrl._depiction_control.plane_thickness_slider.value()
        == layer.plane.thickness
    )


def test_auto_contrast_buttons(qtbot):
    layer = Image(np.arange(8**3).reshape(8, 8, 8), contrast_limits=(0, 1))
    qtctrl = QtImageControls(layer)
    qtbot.addWidget(qtctrl)
    assert layer.contrast_limits == [0, 1]
    qtctrl._contrast_limits_control.auto_scale_bar._once_btn.click()
    assert layer.contrast_limits == [0, 63]

    # change slice
    dims = Dims(
        ndim=3, range=((0, 4, 1), (0, 8, 1), (0, 8, 1)), point=(1, 8, 8)
    )
    layer._slice_dims(dims)
    # hasn't changed yet
    assert layer.contrast_limits == [0, 63]

    # with auto_btn, it should always change
    qtctrl._contrast_limits_control.auto_scale_bar._auto_btn.click()
    assert layer.contrast_limits == [64, 127]
    dims.point = (2, 8, 8)
    layer._slice_dims(dims)
    assert layer.contrast_limits == [128, 191]
    dims.point = (3, 8, 8)
    layer._slice_dims(dims)
    assert layer.contrast_limits == [192, 255]

    # once button turns off continuous
    qtctrl._contrast_limits_control.auto_scale_bar._once_btn.click()
    dims.point = (4, 8, 8)
    layer._slice_dims(dims)
    assert layer.contrast_limits == [192, 255]



from napari._qt.layer_controls.widgets.qt_colormap_control import (
    QtColormapControl,
)
from napari._qt.layer_controls.widgets.qt_contrast_limits import (
    QtContrastLimitsControl,
)
from napari._qt.layer_controls.widgets.qt_face_color import QtFaceColorControl
from napari._qt.layer_controls.widgets.qt_gamma_slider import (
    QtGammaSliderControl,
)
from napari._qt.layer_controls.widgets.qt_opacity_blending_controls import (
    QtOpacityBlendingControls,
)
from napari._qt.layer_controls.widgets.qt_out_slice_checkbox import (
    QtOutSliceCheckBoxControl,
)
from napari._qt.layer_controls.widgets.qt_projection_mode_control import (
    QtProjectionModeControl,
)
from napari._qt.layer_controls.widgets.qt_text_visibility import (
    QtTextVisibilityControl,
)
from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
)

__all__ = [
    'QtColormapControl',
    'QtContrastLimitsControl',
    'QtFaceColorControl',
    'QtGammaSliderControl',
    'QtInterpolationComboBoxControl',
    'QtOpacityBlendingControls',
    'QtOutSliceCheckBoxControl',
    'QtProjectionModeControl',
    'QtTextVisibilityControl',
    'QtWidgetControlsBase',
]



from qtpy.QtWidgets import QComboBox, QWidget

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import qt_signals_blocked
from napari.layers import Image, Points, Vectors
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtProjectionModeControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the layer projection
    mode attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Image | napari.layers.Points | napari.layers.Vectors
        An instance of an Image, Points or Vectors napari layer.

    Attributes
    ----------
    projection_combobox : qtpy.QtWidgets.QComboBox
        ComboBox controlling current projection mode of the layer.
    projection_combobox_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the projection mode chooser widget.
    """

    def __init__(
        self, parent: QWidget, layer: Image | Points | Vectors
    ) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.projection_mode.connect(
            self._on_projection_mode_change
        )

        # Setup widgets
        proj_modes = [i.value for i in self._layer._projectionclass]
        self.projection_combobox = QComboBox(parent)
        self.projection_combobox.addItems(proj_modes)
        connect_setattr(
            self.projection_combobox.currentTextChanged,
            self._layer,
            'projection_mode',
        )

        self._on_projection_mode_change()

        self.projection_combobox_label = QtWrappedLabel(
            trans._('projection mode:')
        )

    def _on_projection_mode_change(self) -> None:
        with qt_signals_blocked(self.projection_combobox):
            self.projection_combobox.setCurrentText(
                str(self._layer.projection_mode)
            )

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.projection_combobox_label, self.projection_combobox)]



import sys

import numpy as np
import numpy.testing as npt
import pytest
from vispy.geometry import PolygonData

from napari.layers.shapes._shapes_models import (
    Ellipse,
    Line,
    Path,
    Polygon,
    Rectangle,
)
from napari.layers.shapes._shapes_utils import (
    triangulate_face,
    triangulate_face_triangle,
)
from napari.settings import get_settings
from napari.settings._experimental import (
    TriangulationBackend,
)

BETTER_TRIANGULATION = 'triangle' in sys.modules or 'bermuda' in sys.modules

BACKEND_TO_MODULE = {
    TriangulationBackend.triangle: 'triangle',
    TriangulationBackend.bermuda: 'bermuda',
    TriangulationBackend.partsegcore: 'PartSegCore_compiled_backend',
    TriangulationBackend.pure_python: 'napari',
}


@pytest.fixture(autouse=True, params=BACKEND_TO_MODULE)
def triangulation_backend(request):
    """Fixture to switch between triangulation backends."""
    settings = get_settings()
    prev = settings.experimental.triangulation_backend
    if BACKEND_TO_MODULE[request.param] not in sys.modules:
        pytest.importorskip(str(request.param))
    settings.experimental.triangulation_backend = request.param
    yield request.param
    settings.experimental.triangulation_backend = prev


def test_rectangle1():
    """Test creating Rectangle by four corners."""
    np.random.seed(0)
    data = np.array([(10, 10), (20, 10), (20, 20), (10, 20)], dtype=np.float32)
    shape = Rectangle(data)
    np.testing.assert_array_equal(shape.data, data)
    assert shape.data_displayed.shape == (4, 2)
    assert shape.slice_key.shape == (2, 0)


def test_rectangle2():
    """Test creating Rectangle by upper left and bottom right."""
    # If given two corners, representation will be expanded to four
    data = np.array([[-10, -10], [20, 20]], dtype=np.float32)
    shape = Rectangle(data)
    assert len(shape.data) == 4
    assert shape.data_displayed.shape == (4, 2)
    assert shape.slice_key.shape == (2, 0)


def test_rectangle_bounding_box():
    """Test that the bounding box is correctly updated based on edge width."""
    data = [[10, 10], [20, 20]]
    shape = Rectangle(data)
    npt.assert_array_equal(
        shape.bounding_box, np.array([[9.5, 9.5], [20.5, 20.5]])
    )
    shape.edge_width = 2
    npt.assert_array_equal(shape.bounding_box, np.array([[9, 9], [21, 21]]))
    shape.edge_width = 4
    npt.assert_array_equal(shape.bounding_box, np.array([[8, 8], [22, 22]]))


def test_rectangle_shift():
    shape = Rectangle(np.array([[0, 0], [1, 0], [1, 1], [0, 1]]))
    npt.assert_array_equal(
        shape.bounding_box, np.array([[-0.5, -0.5], [1.5, 1.5]])
    )

    shape.shift((1, 1))
    npt.assert_array_equal(
        shape.data, np.array([[1, 1], [2, 1], [2, 2], [1, 2]])
    )
    npt.assert_array_equal(
        shape.bounding_box, np.array([[0.5, 0.5], [2.5, 2.5]])
    )


def test_rectangle_rotate():
    shape = Rectangle(np.array([[1, 2], [-1, 2], [-1, -2], [1, -2]]))
    npt.assert_array_equal(
        shape.bounding_box, np.array([[-1.5, -2.5], [1.5, 2.5]])
    )
    shape.rotate(-90)
    npt.assert_array_almost_equal(
        shape.data, np.array([[-2, 1], [-2, -1], [2, -1], [2, 1]])
    )
    npt.assert_array_almost_equal(
        shape.bounding_box, np.array([[-2.5, -1.5], [2.5, 1.5]])
    )


def test_nD_rectangle():
    """Test creating Shape with a single four corner planar 3D rectangle."""
    data = np.array(
        [[0, -10, -10], [0, -10, 20], [0, 20, 20], [0, 20, -10]],
        dtype=np.float32,
    )
    shape = Rectangle(data)
    np.testing.assert_array_equal(shape.data, data)
    assert shape.data_displayed.shape == (4, 2)
    assert shape.slice_key.shape == (2, 1)

    shape.ndisplay = 3
    assert shape.data_displayed.shape == (4, 3)


def test_polygon_data_triangle():
    data = np.array(
        [
            [10.97627008, 14.30378733],
            [12.05526752, 10.89766366],
            [8.47309599, 12.91788226],
            [8.75174423, 17.83546002],
            [19.27325521, 7.66883038],
            [15.83450076, 10.5778984],
        ]
    )
    vertices, _triangles = PolygonData(vertices=data).triangulate()

    assert vertices.shape == (8, 2)


def test_polygon_data_triangle_module():
    pytest.importorskip('triangle')
    data = np.array(
        [
            [10.97627008, 14.30378733],
            [12.05526752, 10.89766366],
            [8.47309599, 12.91788226],
            [8.75174423, 17.83546002],
            [19.27325521, 7.66883038],
            [15.83450076, 10.5778984],
        ]
    )
    vertices, _triangles = triangulate_face(data, triangulate_face_triangle)

    assert vertices.shape == (6, 2)


def test_polygon(triangulation_backend):
    """Test creating Shape with a random polygon."""
    # Test a single non convex six vertex polygon
    data = np.array(
        [
            [10.97627008, 14.30378733],
            [12.05526752, 10.89766366],
            [8.47309599, 12.91788226],
            [8.75174423, 17.83546002],
            [19.27325521, 7.66883038],
            [15.83450076, 10.5778984],
        ],
        dtype=np.float32,
    )
    shape = Polygon(data)
    np.testing.assert_array_equal(shape.data, data)
    assert shape.data_displayed.shape == (6, 2)
    assert shape.slice_key.shape == (2, 0)
    # should get few triangles
    expected_face = (
        (8, 2)
        if triangulation_backend == TriangulationBackend.pure_python
        else (6, 2)
    )
    assert shape._edge_vertices.shape == (16, 2)
    assert shape._face_vertices.shape == expected_face


def test_polygon2():
    data = np.array([[0, 0], [0, 1], [1, 1], [1, 0]])
    shape = Polygon(data, interpolation_order=3)
    # should get many triangles

    expected_face = (249, 2)

    assert shape._edge_vertices.shape == (500, 2)
    assert shape._face_vertices.shape == expected_face


def test_polygon3():
    data = np.array([[0, 0, 0], [0, 0, 1], [0, 1, 1], [1, 1, 1]])
    shape = Polygon(data, interpolation_order=3, ndisplay=3)
    # should get many vertices
    assert shape._edge_vertices.shape == (2500, 3)
    # faces are not made for non-coplanar 3d stuff
    assert shape._face_vertices.shape == (0, 3)


def test_nD_polygon():
    """Test creating Shape with a random nD polygon."""
    # Test a single six vertex planar 3D polygon
    np.random.seed(0)
    data = 20 * np.random.random((6, 3)).astype(np.float32)
    data[:, 0] = 0
    shape = Polygon(data)
    np.testing.assert_array_equal(shape.data, data)
    assert shape.data_displayed.shape == (6, 2)
    assert shape.slice_key.shape == (2, 1)

    shape.ndisplay = 3
    assert shape.data_displayed.shape == (6, 3)


def test_path():
    """Test creating Shape with a random path."""
    # Test a single six vertex path
    np.random.seed(0)
    data = 20 * np.random.random((6, 2)).astype(np.float32)
    shape = Path(data)
    np.testing.assert_array_equal(shape.data, data)
    assert shape.data_displayed.shape == (6, 2)
    assert shape.slice_key.shape == (2, 0)


def test_nD_path():
    """Test creating Shape with a random nD path."""
    # Test a single six vertex 3D path
    np.random.seed(0)
    data = 20 * np.random.random((6, 3)).astype(np.float32)
    shape = Path(data)
    np.testing.assert_array_equal(shape.data, data)
    assert shape.data_displayed.shape == (6, 2)
    assert shape.slice_key.shape == (2, 1)

    shape.ndisplay = 3
    assert shape.data_displayed.shape == (6, 3)


def test_line():
    """Test creating 2D Line."""
    np.random.seed(0)
    data = np.array([[10, 10], [20, 20]], dtype=np.float32)
    shape = Line(data)
    np.testing.assert_array_equal(shape.data, data)
    assert shape.data_displayed.shape == (2, 2)
    assert shape.slice_key.shape == (2, 0)


def test_nD_line():
    """Test creating Line in 3d"""
    data = np.array([[10, 10, 10], [20, 20, 20]], dtype=np.float32)
    shape = Line(data)
    np.testing.assert_array_equal(shape.data, data)
    assert shape.data_displayed.shape == (2, 2)
    assert shape.slice_key.shape == (2, 1)

    shape.ndisplay = 3
    assert shape.data_displayed.shape == (2, 3)


def test_ellipse1():
    """Test creating Ellipse by four corners."""
    data = np.array([(10, 10), (20, 10), (20, 20), (10, 20)], dtype=np.float32)
    shape = Ellipse(data)
    np.testing.assert_array_equal(shape.data, data)
    assert shape.data_displayed.shape == (4, 2)
    assert shape.slice_key.shape == (2, 0)


def test_ellipse2():
    """Test creating Ellipse by upper left and lower right corners."""
    data = np.array([[10, 10], [20, 20]], dtype=np.float32)
    shape = Ellipse(data)
    assert len(shape.data) == 4
    assert shape.data_displayed.shape == (4, 2)
    assert shape.slice_key.shape == (2, 0)


def test_nD_ellipse():
    """Test creating Shape with a random nD ellipse."""
    # Test a single four corner planar 3D ellipse
    data = np.array(
        [(0, -10, -10), (0, -10, 20), (0, 20, 20), (0, 20, -10)],
        dtype=np.float32,
    )
    shape = Ellipse(data)
    np.testing.assert_array_equal(shape.data, data)
    assert shape.data_displayed.shape == (4, 2)
    assert shape.slice_key.shape == (2, 1)

    shape.ndisplay = 3
    assert shape.data_displayed.shape == (4, 3)


def test_ellipse_shift():
    shape = Ellipse(np.array([[0, 0], [1, 0], [1, 1], [0, 1]]))
    npt.assert_array_equal(
        shape.bounding_box, np.array([[-0.5, -0.5], [1.5, 1.5]])
    )

    shape.shift((1, 1))
    npt.assert_array_equal(
        shape.data, np.array([[1, 1], [2, 1], [2, 2], [1, 2]])
    )
    npt.assert_array_equal(
        shape.bounding_box, np.array([[0.5, 0.5], [2.5, 2.5]])
    )


def test_ellipse_rotate():
    shape = Ellipse(np.array([[1, 2], [-1, 2], [-1, -2], [1, -2]]))
    npt.assert_array_equal(
        shape.bounding_box, np.array([[-1.5, -2.5], [1.5, 2.5]])
    )
    shape.rotate(-90)
    npt.assert_array_almost_equal(
        shape.data, np.array([[-2, 1], [-2, -1], [2, -1], [2, 1]])
    )
    npt.assert_array_almost_equal(
        shape.bounding_box, np.array([[-2.5, -1.5], [2.5, 1.5]])
    )



from napari._qt.layer_controls.widgets._vectors.qt_edge_color import (
    QtEdgeColorFeatureControl,
)
from napari._qt.layer_controls.widgets._vectors.qt_line_dimension_spinbox import (
    QtLengthSpinBoxControl,
    QtWidthSpinBoxControl,
)
from napari._qt.layer_controls.widgets._vectors.qt_vector_style_combobox import (
    QtVectorStyleComboBoxControl,
)

__all__ = [
    'QtEdgeColorFeatureControl',
    'QtLengthSpinBoxControl',
    'QtVectorStyleComboBoxControl',
    'QtWidthSpinBoxControl',
]



from typing import TYPE_CHECKING

from napari._qt.layer_controls.qt_layer_controls_base import QtLayerControls
from napari._qt.layer_controls.widgets._labels import (
    QtBrushSizeSliderControl,
    QtColorModeComboBoxControl,
    QtContiguousCheckBoxControl,
    QtContourSpinBoxControl,
    QtDisplaySelectedLabelCheckBoxControl,
    QtLabelControl,
    QtLabelRenderControl,
    QtNdimSpinBoxControl,
    QtPreserveLabelsCheckBoxControl,
)
from napari._qt.utils import set_widgets_enabled_with_opacity
from napari._qt.widgets.qt_mode_buttons import QtModePushButton
from napari.layers.labels._labels_constants import Mode
from napari.utils.translations import trans

if TYPE_CHECKING:
    import napari.layers


INT32_MAX = 2**31 - 1


class QtLabelsControls(QtLayerControls):
    """Qt view and controls for the napari Labels layer.

    Parameters
    ----------
    layer : napari.layers.Labels
        An instance of a napari Labels layer.

    Attributes
    ----------
    _brush_size_slider_control : napari._qt.layer_controls.widgets._labels.QtBrushSizeSliderControl
        Widget that wraps the slider controlling current layer brush size.
    _colormode_combobox_control : napari._qt.layer_controls.widgets._labels.QtColorModeComboBoxControl
        Widget that wraps the comboBox controlling current color mode of the layer.
    _contiguous_checkbox_control : napari._qt.layer_controls.widgets._labels.QtContiguousCheckBoxControl
        Widget that wraps the checkbox to control if label layer fill in is contiguous.
    _contour_spinbox_control : napari._qt.layer_controls.widgets._labels.QtContourSpinBoxControl
        Widget that wraps the spinbox to control the layer contour thickness.
    _display_selected_label_checkbox_control : napari._qt.layer_controls.widgets._labels.QtDisplaySelectedLabelCheckBoxControl
        Widget that wraps the checkbox to control if only the current selected label is shown.
    _ndim_spinbox_control : napari._qt.layer_controls.widgets._labels.QtNdimSpinBoxControl
        Widget that wraps the spinbox to control the number of editable layer dimensions.
    _label_control : napari._qt.layer_controls.widgets._labels.QtLabelControl
        Wrapper widget to handle current label selection.
    _preserve_labels_checkbox_control :napari._qt.layer_controls.widgets._labels.QtPreserveLabelsCheckBoxControl
        Widget that wraps the checkbox to control if existing labels are preserved.
    _render_control : napari._qt.layer_controls.widgets._labels.QtLabelRenderControl
        Wrapper widget to control current label render method.
    colormap_update : napari._qt.widgets.qt_mode_buttons.QtModePushButton
        Button to update colormap of label layer.
    erase_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to select ERASE mode on Labels layer.
    fill_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to select FILL mode on Labels layer.
    paint_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to select PAINT mode on Labels layer.
    pick_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to select PICK mode on Labels layer.
    polygon_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to select POLYGON mode on Labels layer.

    Raises
    ------
    ValueError
        Raise error if label mode is not PAN_ZOOM, PICKER, PAINT, ERASE, or
        FILL.
    """

    layer: 'napari.layers.Labels'
    MODE = Mode
    PAN_ZOOM_ACTION_NAME = 'activate_labels_pan_zoom_mode'
    TRANSFORM_ACTION_NAME = 'activate_labels_transform_mode'

    def __init__(self, layer) -> None:
        super().__init__(layer)

        # Setup buttons
        # shuffle colormap button
        self.colormap_update = QtModePushButton(
            layer,
            'shuffle',
            slot=self.changeColor,
            tooltip=trans._('Shuffle colors'),
        )

        self.pick_button = self._radio_button(
            layer,
            'picker',
            Mode.PICK,
            True,
            'activate_labels_picker_mode',
        )
        self.paint_button = self._radio_button(
            layer,
            'paint',
            Mode.PAINT,
            True,
            'activate_labels_paint_mode',
        )
        self.polygon_button = self._radio_button(
            layer,
            'labels_polygon',
            Mode.POLYGON,
            True,
            'activate_labels_polygon_mode',
        )
        self.fill_button = self._radio_button(
            layer,
            'fill',
            Mode.FILL,
            True,
            'activate_labels_fill_mode',
        )
        self.erase_button = self._radio_button(
            layer,
            'erase',
            Mode.ERASE,
            True,
            'activate_labels_erase_mode',
        )
        # don't bind with action manager as this would remove "Toggle with {shortcut}"
        self._on_editable_or_visible_change()

        self.button_grid.addWidget(self.colormap_update, 0, 0)
        self.button_grid.addWidget(self.erase_button, 0, 1)
        self.button_grid.addWidget(self.paint_button, 0, 2)
        self.button_grid.addWidget(self.polygon_button, 0, 3)
        self.button_grid.addWidget(self.fill_button, 0, 4)
        self.button_grid.addWidget(self.pick_button, 0, 5)

        # Setup widgets controls
        self._label_control = QtLabelControl(self, layer)
        self._add_widget_controls(self._label_control)
        self._brush_size_slider_control = QtBrushSizeSliderControl(self, layer)
        self._add_widget_controls(self._brush_size_slider_control)
        self._render_control = QtLabelRenderControl(self, layer)
        self._add_widget_controls(self._render_control)
        self._colormode_combobox_control = QtColorModeComboBoxControl(
            self, layer
        )
        self._add_widget_controls(self._colormode_combobox_control)
        self._contour_spinbox_control = QtContourSpinBoxControl(self, layer)
        self._add_widget_controls(self._contour_spinbox_control)
        self._ndim_spinbox_control = QtNdimSpinBoxControl(self, layer)
        self._add_widget_controls(self._ndim_spinbox_control)
        self._contiguous_checkbox_control = QtContiguousCheckBoxControl(
            self, layer
        )
        self._add_widget_controls(self._contiguous_checkbox_control)
        self._preserve_labels_checkbox_control = (
            QtPreserveLabelsCheckBoxControl(self, layer)
        )
        self._add_widget_controls(self._preserve_labels_checkbox_control)
        self._display_selected_label_checkbox_control = (
            QtDisplaySelectedLabelCheckBoxControl(self, layer)
        )
        self._add_widget_controls(
            self._display_selected_label_checkbox_control
        )

        self._on_ndisplay_changed()

    def _on_mode_change(self, event):
        """Receive layer model mode change event and update checkbox ticks.

        Parameters
        ----------
        event : napari.utils.event.Event
            The napari event that triggered this method.

        Raises
        ------
        ValueError
            Raise error if event.mode is not PAN_ZOOM, PICK, PAINT, ERASE, FILL
            or TRANSFORM
        """
        super()._on_mode_change(event)

    def changeColor(self):
        """Change colormap of the label layer."""
        self.layer.new_colormap()

    def _on_editable_or_visible_change(self):
        super()._on_editable_or_visible_change()
        self._set_polygon_tool_state()

    def _on_ndisplay_changed(self):
        show_3d_widgets = self.ndisplay == 3
        if show_3d_widgets:
            self._render_control._on_display_change_show()
        else:
            self._render_control._on_display_change_hide()
        self._on_editable_or_visible_change()
        self._set_polygon_tool_state()
        super()._on_ndisplay_changed()

    def _set_polygon_tool_state(self):
        if hasattr(self, 'polygon_button'):
            set_widgets_enabled_with_opacity(
                self, [self.polygon_button], self._is_polygon_tool_enabled()
            )

    def _is_polygon_tool_enabled(self):
        return (
            self.layer.editable
            and self.layer.visible
            and self.layer.n_edit_dimensions == 2
            and self.ndisplay == 2
        )



import logging
import time
import warnings
from dataclasses import dataclass
from pathlib import Path

from napari.utils.translations import trans


@dataclass
class StartupScriptStatusInfo:
    startup_time: float
    script_path: Path
    script_code: str


startup_script_status_info: StartupScriptStatusInfo | None = None


def _run_configured_startup_script() -> None:
    from napari.settings import get_settings

    logger = logging.getLogger(__name__)

    if not get_settings().application.startup_script:
        logger.debug(
            'No startup script is set in the settings. Skipping execution.'
        )
        return
    script_path = (
        Path(get_settings().application.startup_script).expanduser().resolve()
    )

    if script_path == Path().resolve():
        # empty path
        return

    if not (script_path.exists() and script_path.is_file()):
        warnings.warn(
            trans._(
                'Startup script path is set to {script_path}. This path does not have a valid startup script. Please check the setting. napari will be launched without a startup script.',
                deferred=True,
                script_path=script_path,
            )
        )
        return

    from napari_builtins.io._read import (
        execute_python_code,
    )

    script_code = script_path.read_text()
    start_time = time.time()

    execute_python_code(script_code, script_path)

    total_time = time.time() - start_time

    logger.info(
        'Loaded startup script: %s in %f seconds', script_path, total_time
    )

    globals()['startup_script_status_info'] = StartupScriptStatusInfo(
        startup_time=total_time,
        script_path=script_path,
        script_code=script_code,
    )



import contextlib

import numpy as np
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QWidget,
)
from superqt import QLabeledSlider

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import qt_signals_blocked
from napari.layers import Points
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtCurrentSizeSliderControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the current
    size layer attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Points
        An instance of a napari Points layer.

    Attributes
    ----------
    size_slider : superqt.QLabeledDoubleSlider
        Slider controlling current size attribute of the layer.
    size_slider_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the size chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Points) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.size.connect(self._on_current_size_change)
        self._layer.events.current_size.connect(self._on_current_size_change)

        # Setup widgets
        sld = QLabeledSlider(Qt.Orientation.Horizontal)
        sld.setToolTip(
            trans._(
                'Change the size of currently selected points and any added afterwards.'
            )
        )
        sld.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        sld.setMinimum(1)
        if self._layer.size.size:
            max_value = max(100, int(np.max(self._layer.size)) + 1)
        else:
            max_value = 100
        sld.setMaximum(max_value)
        sld.setSingleStep(1)
        value = self._layer.current_size
        sld.setValue(int(value))
        self.size_slider = sld
        connect_setattr(
            self.size_slider.valueChanged, self._layer, 'current_size'
        )

        self.size_slider_label = QtWrappedLabel(trans._('point size:'))

    def _on_current_size_change(self) -> None:
        """Receive layer model size change event and update point size slider."""
        with qt_signals_blocked(self.size_slider):
            value = self._layer.current_size
            min_val = min(value) if isinstance(value, list) else value
            max_val = max(value) if isinstance(value, list) else value
            if min_val < self.size_slider.minimum():
                self.size_slider.setMinimum(max(1, int(min_val - 1)))
            if max_val > self.size_slider.maximum():
                self.size_slider.setMaximum(int(max_val + 1))
            with contextlib.suppress(TypeError):
                self.size_slider.setValue(int(value))

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.size_slider_label, self.size_slider)]



from unittest.mock import Mock

import numpy as np
import numpy.testing as npt
import pint
import pytest

from napari.layers.base._test_util_sample_layer import SampleLayer

REG = pint.get_application_registry()


def test_assign_units():
    layer = SampleLayer(np.empty((10, 10)))
    mock = Mock()
    layer.events.units.connect(mock)
    assert layer.units == (REG.pixel, REG.pixel)

    layer.units = ('nm', 'nm')
    mock.assert_called_once()
    mock.reset_mock()

    assert layer.units == (REG.nm, REG.nm)

    layer.units = (REG.mm, REG.mm)
    mock.assert_called_once()
    mock.reset_mock()

    assert layer.units == (REG.mm, REG.mm)

    layer.units = ('mm', 'mm')
    mock.assert_not_called()

    layer.units = 'km'
    mock.assert_called_once()
    mock.reset_mock()
    assert layer.units == (REG.km, REG.km)

    layer.units = None
    mock.assert_called_once()
    assert layer.units == (REG.pixel, REG.pixel)


def test_units_constructor():
    layer = SampleLayer(np.empty((10, 10)), units=('nm', 'nm'))
    assert layer.units == (REG.nm, REG.nm)

    layer = SampleLayer(np.empty((10, 10)), units=(REG.mm, REG.mm))
    assert layer.units == (REG.mm, REG.mm)

    layer = SampleLayer(np.empty((10, 10)), units=('mm', 'mm'))
    assert layer.units == (REG.mm, REG.mm)

    layer = SampleLayer(np.empty((10, 10)), units=None)
    assert layer.units == (REG.pixel, REG.pixel)


def test_assign_units_error():
    layer = SampleLayer(np.empty((10, 10)))
    with pytest.raises(ValueError, match='must have length ndim'):
        layer.units = ('m', 'm', 'm')

    with pytest.raises(ValueError, match='Could not find unit'):
        layer.units = ('ugh', 'ugh')

    with pytest.raises(ValueError, match='Could not find unit'):
        SampleLayer(np.empty((10, 10)), units=('ugh', 'ugh'))

    with pytest.raises(ValueError, match='must have length ndim'):
        SampleLayer(np.empty((10, 10)), units=('m', 'm', 'm'))


def test_axis_labels_assign():
    layer = SampleLayer(np.empty((10, 10)))
    mock = Mock()
    layer.events.axis_labels.connect(mock)
    assert layer.axis_labels == ('axis -2', 'axis -1')

    layer.axis_labels = ('x', 'y')
    mock.assert_called_once()
    mock.reset_mock()

    assert layer.axis_labels == ('x', 'y')

    layer.axis_labels = ('x', 'y')
    mock.assert_not_called()

    layer.axis_labels = None
    mock.assert_called_once()
    assert layer.axis_labels == ('axis -2', 'axis -1')


def test_axis_labels_constructor():
    layer = SampleLayer(np.empty((10, 10)), axis_labels=('x', 'y'))
    assert layer.axis_labels == ('x', 'y')

    layer = SampleLayer(np.empty((10, 10)), axis_labels=None)
    assert layer.axis_labels == ('axis -2', 'axis -1')


def test_axis_labels_error():
    layer = SampleLayer(np.empty((10, 10)))
    with pytest.raises(ValueError, match='must have length ndim'):
        layer.axis_labels = ('x', 'y', 'z')

    with pytest.raises(ValueError, match='must have length ndim'):
        SampleLayer(np.empty((10, 10)), axis_labels=('x', 'y', 'z'))


def test_non_visible_mode():
    layer = SampleLayer(np.empty((10, 10)))
    layer.mode = 'transform'

    # change layer visibility and check the layer mode gets updated
    layer.visible = False
    assert layer.mode == 'pan_zoom'
    layer.visible = True
    assert layer.mode == 'transform'


def test_world_to_displayed_data_normal_3D():
    layer = SampleLayer(np.empty((10, 10, 10)))
    layer.scale = (1, 3, 2)

    normal_vector = [0, 1, 1]

    expected_transformed_vector = [0, 3 * (13**0.5) / 13, 2 * (13**0.5) / 13]

    transformed_vector = layer._world_to_displayed_data_normal(
        normal_vector, dims_displayed=[0, 1, 2]
    )

    assert np.allclose(transformed_vector, expected_transformed_vector)


def test_world_to_displayed_data_normal_4D():
    layer = SampleLayer(np.empty((10, 10, 10, 10)))
    layer.scale = (1, 3, 2, 1)

    normal_vector = [0, 1, 1]

    expected_transformed_vector = [0, 3 * (13**0.5) / 13, 2 * (13**0.5) / 13]

    transformed_vector = layer._world_to_displayed_data_normal(
        normal_vector, dims_displayed=[0, 1, 2]
    )

    assert np.allclose(transformed_vector, expected_transformed_vector)


def test_invalidate_extent_scale():
    """Test that the extent is invalidated when the data changes."""
    layer = SampleLayer(np.empty((10, 10)))
    npt.assert_array_equal(layer.extent.step, (1, 1))
    with layer._block_refresh():
        layer.scale = (2, 2)
    npt.assert_array_equal(layer.extent.step, (2, 2))


def test_invalidate_extent_units():
    """Test that the extent is invalidated when the data changes."""
    # commented lines required 7889 and should be uncomment later
    layer = SampleLayer(np.empty((10, 10)))
    # px = pint.get_application_registry().pixel
    # mm = pint.get_application_registry().mm
    # npt.assert_array_equal(layer.extent.units, (px, px))
    with layer._block_refresh():
        layer.units = ('mm', 'mm')
    # npt.assert_array_equal(layer.extent.units, (mm, mm))


def test_invalidate_extent_translate():
    """Test that the extent is invalidated when the data changes."""
    layer = SampleLayer(np.empty((10, 10)))
    npt.assert_array_equal(layer.extent.world[0], (0, 0))
    with layer._block_refresh():
        layer.translate = (1, 1)
    npt.assert_array_equal(layer.extent.world[0], (1, 1))


def test_invalidate_extent_rotate():
    """Test that the extent is invalidated when the data changes."""
    layer = SampleLayer(np.empty((10, 20)))
    npt.assert_array_equal(layer.extent.world, [[0, 0], [9, 19]])
    with layer._block_refresh():
        layer.rotate = 90
    npt.assert_almost_equal(layer.extent.world, [[-19, 0], [0, 9]])


def test_invalidate_extent_affine():
    """Test that the extent is invalidated when the data changes."""
    layer = SampleLayer(np.empty((10, 20)))
    npt.assert_array_equal(layer.extent.world, [[0, 0], [9, 19]])
    with layer._block_refresh():
        layer.affine = [[2, 0, 0], [0, 2, 0], [0, 0, 1]]
    npt.assert_array_equal(layer.extent.world, [[0, 0], [18, 38]])


def test_invalidate_extent_shear():
    """Test that the extent is invalidated when the data changes."""
    layer = SampleLayer(np.empty((10, 20)))
    npt.assert_array_equal(layer.extent.world, [[0, 0], [9, 19]])
    with layer._block_refresh():
        layer.shear = [1]
    npt.assert_array_equal(layer.extent.world, [[0, 0], [28, 19]])
